# -----------------------------------------------------------------------------
# sly: yacc.py
#
# Copyright (C) 2016-2018
# David M. Beazley (Dabeaz LLC)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of the David Beazley or Dabeaz LLC may be used to
#   endorse or promote products derived from this software without
#  specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

import sys
import inspect
from collections import OrderedDict, defaultdict, Counter

__all__        = [ 'Parser' ]

class YaccError(Exception):
    '''
    Exception raised for yacc-related build errors.
    '''
    pass

#-----------------------------------------------------------------------------
#                     === User configurable parameters ===
#
# Change these to modify the default behavior of yacc (if you wish).  
# Move these parameters to the Yacc class itself.
#-----------------------------------------------------------------------------

ERROR_COUNT = 3                # Number of symbols that must be shifted to leave recovery mode
MAXINT = sys.maxsize

# This object is a stand-in for a logging object created by the
# logging module.   SLY will use this by default to create things
# such as the parser.out file.  If a user wants more detailed
# information, they can create their own logging object and pass
# it into SLY.

class SlyLogger(object):
    def __init__(self, f):
        self.f = f

    def debug(self, msg, *args, **kwargs):
        self.f.write((msg % args) + '\n')

    info = debug

    def warning(self, msg, *args, **kwargs):
        self.f.write('WARNING: ' + (msg % args) + '\n')

    def error(self, msg, *args, **kwargs):
        self.f.write('ERROR: ' + (msg % args) + '\n')

    critical = debug


# ----------------------------------------------------------------------
# This class is used to hold non-terminal grammar symbols during parsing.
# It normally has the following attributes set:
#        .type       = Grammar symbol type
#        .value      = Symbol value
#        .lineno     = Starting line number
#        .index      = Starting lex position
# ----------------------------------------------------------------------

class YaccSymbol:
    def __str__(self):
        return self.type

    def __repr__(self):
        return str(self)

# ----------------------------------------------------------------------
# This class is a wrapper around the objects actually passed to each
# grammar rule.   Index lookup and assignment actually assign the
# .value attribute of the underlying YaccSymbol object.
# The lineno() method returns the line number of a given
# item (or 0 if not defined).   
# ----------------------------------------------------------------------

class YaccProduction:
    __slots__ = ('_slice', '_namemap', '_stack')
    def __init__(self, s, stack=None):
        self._slice = s
        self._namemap = { }
        self._stack = stack

    def __getitem__(self, n):
        if n >= 0:
            return self._slice[n].value
        else:
            return self._stack[n].value

    def __setitem__(self, n, v):
        if n >= 0:
            self._slice[n].value = v
        else:
            self._stack[n].value = v

    def __len__(self):
        return len(self._slice)

    @property
    def lineno(self):
        for tok in self._slice:
            lineno = getattr(tok, 'lineno', None)
            if lineno:
                return lineno
        raise AttributeError('No line number found')

    @property
    def index(self):
        for tok in self._slice:
            index = getattr(tok, 'index', None)
            if index is not None:
                return index
        raise AttributeError('No index attribute found')

    @property
    def end(self):
        result = None
        for tok in self._slice:
            r = getattr(tok, 'end', None)
            if r:
                result = r
        return result
    
    def __getattr__(self, name):
        if name in self._namemap:
            return self._namemap[name](self._slice)
        else:
            nameset = '{' + ', '.join(self._namemap) + '}'
            raise AttributeError(f'No symbol {name}. Must be one of {nameset}.')

    def __setattr__(self, name, value):
        if name[:1] == '_':
            super().__setattr__(name, value)
        else:
            raise AttributeError(f"Can't reassign the value of attribute {name!r}")

# -----------------------------------------------------------------------------
#                          === Grammar Representation ===
#
# The following functions, classes, and variables are used to represent and
# manipulate the rules that make up a grammar.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# class Production:
#
# This class stores the raw information about a single production or grammar rule.
# A grammar rule refers to a specification such as this:
#
#       expr : expr PLUS term
#
# Here are the basic attributes defined on all productions
#
#       name     - Name of the production.  For example 'expr'
#       prod     - A list of symbols on the right side ['expr','PLUS','term']
#       prec     - Production precedence level
#       number   - Production number.
#       func     - Function that executes on reduce
#       file     - File where production function is defined
#       lineno   - Line number where production function is defined
#
# The following attributes are defined or optional.
#
#       len       - Length of the production (number of symbols on right hand side)
#       usyms     - Set of unique symbols found in the production
# -----------------------------------------------------------------------------

class Production(object):
    reduced = 0
    def __init__(self, number, name, prod, precedence=('right', 0), func=None, file='', line=0):
        self.name     = name
        self.prod     = tuple(prod)
        self.number   = number
        self.func     = func
        self.file     = file
        self.line     = line
        self.prec     = precedence
        
        # Internal settings used during table construction
        self.len  = len(self.prod)   # Length of the production

        # Create a list of unique production symbols used in the production
        self.usyms = []
        symmap = defaultdict(list)
        for n, s in enumerate(self.prod):
            symmap[s].append(n)
            if s not in self.usyms:
                self.usyms.append(s)

        # Create a name mapping
        # First determine (in advance) if there are duplicate names
        namecount = defaultdict(int)
        for key in self.prod:
            namecount[key] += 1
            if key in _name_aliases:
                for key in _name_aliases[key]:
                    namecount[key] += 1

        # Now, walk through the names and generate accessor functions
        nameuse = defaultdict(int)
        namemap = { }
        for index, key in enumerate(self.prod):
            if namecount[key] > 1:
                k = f'{key}{nameuse[key]}'
                nameuse[key] += 1
            else:
                k = key
            namemap[k] = lambda s,i=index: s[i].value
            if key in _name_aliases:
                for n, alias in enumerate(_name_aliases[key]):
                    if namecount[alias] > 1:
                        k = f'{alias}{nameuse[alias]}'
                        nameuse[alias] += 1
                    else:
                        k = alias
                    # The value is either a list (for repetition) or a tuple for optional 
                    namemap[k] = lambda s,i=index,n=n: ([x[n] for x in s[i].value]) if isinstance(s[i].value, list) else s[i].value[n]

        self.namemap = namemap
                
        # List of all LR items for the production
        self.lr_items = []
        self.lr_next = None

    def __str__(self):
        if self.prod:
            s = '%s -> %s' % (self.name, ' '.join(self.prod))
        else:
            s = f'{self.name} -> <empty>'

        if self.prec[1]:
            s += '  [precedence=%s, level=%d]' % self.prec

        return s

    def __repr__(self):
        return f'Production({self})'

    def __len__(self):
        return len(self.prod)

    def __nonzero__(self):
        raise RuntimeError('Used')
        return 1

    def __getitem__(self, index):
        return self.prod[index]

    # Return the nth lr_item from the production (or None if at the end)
    def lr_item(self, n):
        if n > len(self.prod):
            return None
        p = LRItem(self, n)
        # Precompute the list of productions immediately following.
        try:
            p.lr_after = Prodnames[p.prod[n+1]]
        except (IndexError, KeyError):
            p.lr_after = []
        try:
            p.lr_before = p.prod[n-1]
        except IndexError:
            p.lr_before = None
        return p

# -----------------------------------------------------------------------------
# class LRItem
#
# This class represents a specific stage of parsing a production rule.  For
# example:
#
#       expr : expr . PLUS term
#
# In the above, the "." represents the current location of the parse.  Here
# basic attributes:
#
#       name       - Name of the production.  For example 'expr'
#       prod       - A list of symbols on the right side ['expr','.', 'PLUS','term']
#       number     - Production number.
#
#       lr_next      Next LR item. Example, if we are ' expr -> expr . PLUS term'
#                    then lr_next refers to 'expr -> expr PLUS . term'
#       lr_index   - LR item index (location of the ".") in the prod list.
#       lookaheads - LALR lookahead symbols for this item
#       len        - Length of the production (number of symbols on right hand side)
#       lr_after    - List of all productions that immediately follow
#       lr_before   - Grammar symbol immediately before
# -----------------------------------------------------------------------------

class LRItem(object):
    def __init__(self, p, n):
        self.name       = p.name
        self.prod       = list(p.prod)
        self.number     = p.number
        self.lr_index   = n
        self.lookaheads = {}
        self.prod.insert(n, '.')
        self.prod       = tuple(self.prod)
        self.len        = len(self.prod)
        self.usyms      = p.usyms

    def __str__(self):
        if self.prod:
            s = '%s -> %s' % (self.name, ' '.join(self.prod))
        else:
            s = f'{self.name} -> <empty>'
        return s

    def __repr__(self):
        return f'LRItem({self})'

# -----------------------------------------------------------------------------
# rightmost_terminal()
#
# Return the rightmost terminal from a list of symbols.  Used in add_production()
# -----------------------------------------------------------------------------
def rightmost_terminal(symbols, terminals):
    i = len(symbols) - 1
    while i >= 0:
        if symbols[i] in terminals:
            return symbols[i]
        i -= 1
    return None

# -----------------------------------------------------------------------------
#                           === GRAMMAR CLASS ===
#
# The following class represents the contents of the specified grammar along
# with various computed properties such as first sets, follow sets, LR items, etc.
# This data is used for critical parts of the table generation process later.
# -----------------------------------------------------------------------------

class GrammarError(YaccError):
    pass

class Grammar(object):
    def __init__(self, terminals):
        self.Productions  = [None]  # A list of all of the productions.  The first
                                    # entry is always reserved for the purpose of
                                    # building an augmented grammar

        self.Prodnames    = {}      # A dictionary mapping the names of nonterminals to a list of all
                                    # productions of that nonterminal.

        self.Prodmap      = {}      # A dictionary that is only used to detect duplicate
                                    # productions.

        self.Terminals    = {}      # A dictionary mapping the names of terminal symbols to a
                                    # list of the rules where they are used.

        for term in terminals:
            self.Terminals[term] = []

        self.Terminals['error'] = []

        self.Nonterminals = {}      # A dictionary mapping names of nonterminals to a list
                                    # of rule numbers where they are used.

        self.First        = {}      # A dictionary of precomputed FIRST(x) symbols

        self.Follow       = {}      # A dictionary of precomputed FOLLOW(x) symbols

        self.Precedence   = {}      # Precedence rules for each terminal. Contains tuples of the
                                    # form ('right',level) or ('nonassoc', level) or ('left',level)

        self.UsedPrecedence = set() # Precedence rules that were actually used by the grammer.
                                    # This is only used to provide error checking and to generate
                                    # a warning about unused precedence rules.

        self.Start = None           # Starting symbol for the grammar


    def __len__(self):
        return len(self.Productions)

    def __getitem__(self, index):
        return self.Productions[index]

    # -----------------------------------------------------------------------------
    # set_precedence()
    #
    # Sets the precedence for a given terminal. assoc is the associativity such as
    # 'left','right', or 'nonassoc'.  level is a numeric level.
    #
    # -----------------------------------------------------------------------------

    def set_precedence(self, term, assoc, level):
        assert self.Productions == [None], 'Must call set_precedence() before add_production()'
        if term in self.Precedence:
            raise GrammarError(f'Precedence already specified for terminal {term!r}')
        if assoc not in ['left', 'right', 'nonassoc']:
            raise GrammarError(f"Associativity of {term!r} must be one of 'left','right', or 'nonassoc'")
        self.Precedence[term] = (assoc, level)

    # -----------------------------------------------------------------------------
    # add_production()
    #
    # Given an action function, this function assembles a production rule and
    # computes its precedence level.
    #
    # The production rule is supplied as a list of symbols.   For example,
    # a rule such as 'expr : expr PLUS term' has a production name of 'expr' and
    # symbols ['expr','PLUS','term'].
    #
    # Precedence is determined by the precedence of the right-most non-terminal
    # or the precedence of a terminal specified by %prec.
    #
    # A variety of error checks are performed to make sure production symbols
    # are valid and that %prec is used correctly.
    # -----------------------------------------------------------------------------

    def add_production(self, prodname, syms, func=None, file='', line=0):

        if prodname in self.Terminals:
            raise GrammarError(f'{file}:{line}: Illegal rule name {prodname!r}. Already defined as a token')
        if prodname == 'error':
            raise GrammarError(f'{file}:{line}: Illegal rule name {prodname!r}. error is a reserved word')

        # Look for literal tokens
        for n, s in enumerate(syms):
            if s[0] in "'\"" and s[0] == s[-1]:
                c = s[1:-1]
                if (len(c) != 1):
                    raise GrammarError(f'{file}:{line}: Literal token {s} in rule {prodname!r} may only be a single character')
                if c not in self.Terminals:
                    self.Terminals[c] = []
                syms[n] = c
                continue

        # Determine the precedence level
        if '%prec' in syms:
            if syms[-1] == '%prec':
                raise GrammarError(f'{file}:{line}: Syntax error. Nothing follows %%prec')
            if syms[-2] != '%prec':
                raise GrammarError(f'{file}:{line}: Syntax error. %prec can only appear at the end of a grammar rule')
            precname = syms[-1]
            prodprec = self.Precedence.get(precname)
            if not prodprec:
                raise GrammarError(f'{file}:{line}: Nothing known about the precedence of {precname!r}')
            else:
                self.UsedPrecedence.add(precname)
            del syms[-2:]     # Drop %prec from the rule
        else:
            # If no %prec, precedence is determined by the rightmost terminal symbol
            precname = rightmost_terminal(syms, self.Terminals)
            prodprec = self.Precedence.get(precname, ('right', 0))

        # See if the rule is already in the rulemap
        map = '%s -> %s' % (prodname, syms)
        if map in self.Prodmap:
            m = self.Prodmap[map]
            raise GrammarError(f'{file}:{line}: Duplicate rule {m}. ' +
                               f'Previous definition at {m.file}:{m.line}')

        # From this point on, everything is valid.  Create a new Production instance
        pnumber  = len(self.Productions)
        if prodname not in self.Nonterminals:
            self.Nonterminals[prodname] = []

        # Add the production number to Terminals and Nonterminals
        for t in syms:
            if t in self.Terminals:
                self.Terminals[t].append(pnumber)
            else:
                if t not in self.Nonterminals:
                    self.Nonterminals[t] = []
                self.Nonterminals[t].append(pnumber)

        # Create a production and add it to the list of productions
        p = Production(pnumber, prodname, syms, prodprec, func, file, line)
        self.Productions.append(p)
        self.Prodmap[map] = p

        # Add to the global productions list
        try:
            self.Prodnames[prodname].append(p)
        except KeyError:
            self.Prodnames[prodname] = [p]

    # -----------------------------------------------------------------------------
    # set_start()
    #
    # Sets the starting symbol and creates the augmented grammar.  Production
    # rule 0 is S' -> start where start is the start symbol.
    # -----------------------------------------------------------------------------

    def set_start(self, start=None):
        if callable(start):
            start = start.__name__

        if not start:
            start = self.Productions[1].name

        if start not in self.Nonterminals:
            raise GrammarError(f'start symbol {start} undefined')
        self.Productions[0] = Production(0, "S'", [start])
        self.Nonterminals[start].append(0)
        self.Start = start

    # -----------------------------------------------------------------------------
    # find_unreachable()
    #
    # Find all of the nonterminal symbols that can't be reached from the starting
    # symbol.  Returns a list of nonterminals that can't be reached.
    # -----------------------------------------------------------------------------

    def find_unreachable(self):

        # Mark all symbols that are reachable from a symbol s
        def mark_reachable_from(s):
            if s in reachable:
                return
            reachable.add(s)
            for p in self.Prodnames.get(s, []):
                for r in p.prod:
                    mark_reachable_from(r)

        reachable = set()
        mark_reachable_from(self.Productions[0].prod[0])
        return [s for s in self.Nonterminals if s not in reachable]

    # -----------------------------------------------------------------------------
    # infinite_cycles()
    #
    # This function looks at the various parsing rules and tries to detect
    # infinite recursion cycles (grammar rules where there is no possible way
    # to derive a string of only terminals).
    # -----------------------------------------------------------------------------

    def infinite_cycles(self):
        terminates = {}

        # Terminals:
        for t in self.Terminals:
            terminates[t] = True

        terminates['$end'] = True

        # Nonterminals:

        # Initialize to false:
        for n in self.Nonterminals:
            terminates[n] = False

        # Then propagate termination until no change:
        while True:
            some_change = False
            for (n, pl) in self.Prodnames.items():
                # Nonterminal n terminates iff any of its productions terminates.
                for p in pl:
                    # Production p terminates iff all of its rhs symbols terminate.
                    for s in p.prod:
                        if not terminates[s]:
                            # The symbol s does not terminate,
                            # so production p does not terminate.
                            p_terminates = False
                            break
                    else:
                        # didn't break from the loop,
                        # so every symbol s terminates
                        # so production p terminates.
                        p_terminates = True

                    if p_terminates:
                        # symbol n terminates!
                        if not terminates[n]:
                            terminates[n] = True
                            some_change = True
                        # Don't need to consider any more productions for this n.
                        break

            if not some_change:
                break

        infinite = []
        for (s, term) in terminates.items():
            if not term:
                if s not in self.Prodnames and s not in self.Terminals and s != 'error':
                    # s is used-but-not-defined, and we've already warned of that,
                    # so it would be overkill to say that it's also non-terminating.
                    pass
                else:
                    infinite.append(s)

        return infinite

    # -----------------------------------------------------------------------------
    # undefined_symbols()
    #
    # Find all symbols that were used the grammar, but not defined as tokens or
    # grammar rules.  Returns a list of tuples (sym, prod) where sym in the symbol
    # and prod is the production where the symbol was used.
    # -----------------------------------------------------------------------------
    def undefined_symbols(self):
        result = []
        for p in self.Productions:
            if not p:
                continue

            for s in p.prod:
                if s not in self.Prodnames and s not in self.Terminals and s != 'error':
                    result.append((s, p))
        return result

    # -----------------------------------------------------------------------------
    # unused_terminals()
    #
    # Find all terminals that were defined, but not used by the grammar.  Returns
    # a list of all symbols.
    # -----------------------------------------------------------------------------
    def unused_terminals(self):
        unused_tok = []
        for s, v in self.Terminals.items():
            if s != 'error' and not v:
                unused_tok.append(s)

        return unused_tok

    # ------------------------------------------------------------------------------
    # unused_rules()
    #
    # Find all grammar rules that were defined,  but not used (maybe not reachable)
    # Returns a list of productions.
    # ------------------------------------------------------------------------------

    def unused_rules(self):
        unused_prod = []
        for s, v in self.Nonterminals.items():
            if not v:
                p = self.Prodnames[s][0]
                unused_prod.append(p)
        return unused_prod

    # -----------------------------------------------------------------------------
    # unused_precedence()
    #
    # Returns a list of tuples (term,precedence) corresponding to precedence
    # rules that were never used by the grammar.  term is the name of the terminal
    # on which precedence was applied and precedence is a string such as 'left' or
    # 'right' corresponding to the type of precedence.
    # -----------------------------------------------------------------------------

    def unused_precedence(self):
        unused = []
        for termname in self.Precedence:
            if not (termname in self.Terminals or termname in self.UsedPrecedence):
                unused.append((termname, self.Precedence[termname][0]))

        return unused

    # -------------------------------------------------------------------------
    # _first()
    #
    # Compute the value of FIRST1(beta) where beta is a tuple of symbols.
    #
    # During execution of compute_first1, the result may be incomplete.
    # Afterward (e.g., when called from compute_follow()), it will be complete.
    # -------------------------------------------------------------------------
    def _first(self, beta):

        # We are computing First(x1,x2,x3,...,xn)
        result = []
        for x in beta:
            x_produces_empty = False

            # Add all the non-<empty> symbols of First[x] to the result.
            for f in self.First[x]:
                if f == '<empty>':
                    x_produces_empty = True
                else:
                    if f not in result:
                        result.append(f)

            if x_produces_empty:
                # We have to consider the next x in beta,
                # i.e. stay in the loop.
                pass
            else:
                # We don't have to consider any further symbols in beta.
                break
        else:
            # There was no 'break' from the loop,
            # so x_produces_empty was true for all x in beta,
            # so beta produces empty as well.
            result.append('<empty>')

        return result

    # -------------------------------------------------------------------------
    # compute_first()
    #
    # Compute the value of FIRST1(X) for all symbols
    # -------------------------------------------------------------------------
    def compute_first(self):
        if self.First:
            return self.First

        # Terminals:
        for t in self.Terminals:
            self.First[t] = [t]

        self.First['$end'] = ['$end']

        # Nonterminals:

        # Initialize to the empty set:
        for n in self.Nonterminals:
            self.First[n] = []

        # Then propagate symbols until no change:
        while True:
            some_change = False
            for n in self.Nonterminals:
                for p in self.Prodnames[n]:
                    for f in self._first(p.prod):
                        if f not in self.First[n]:
                            self.First[n].append(f)
                            some_change = True
            if not some_change:
                break

        return self.First

    # ---------------------------------------------------------------------
    # compute_follow()
    #
    # Computes all of the follow sets for every non-terminal symbol.  The
    # follow set is the set of all symbols that might follow a given
    # non-terminal.  See the Dragon book, 2nd Ed. p. 189.
    # ---------------------------------------------------------------------
    def compute_follow(self, start=None):
        # If already computed, return the result
        if self.Follow:
            return self.Follow

        # If first sets not computed yet, do that first.
        if not self.First:
            self.compute_first()

        # Add '$end' to the follow list of the start symbol
        for k in self.Nonterminals:
            self.Follow[k] = []

        if not start:
            start = self.Productions[1].name

        self.Follow[start] = ['$end']

        while True:
            didadd = False
            for p in self.Productions[1:]:
                # Here is the production set
                for i, B in enumerate(p.prod):
                    if B in self.Nonterminals:
                        # Okay. We got a non-terminal in a production
                        fst = self._first(p.prod[i+1:])
                        hasempty = False
                        for f in fst:
                            if f != '<empty>' and f not in self.Follow[B]:
                                self.Follow[B].append(f)
                                didadd = True
                            if f == '<empty>':
                                hasempty = True
                        if hasempty or i == (len(p.prod)-1):
                            # Add elements of follow(a) to follow(b)
                            for f in self.Follow[p.name]:
                                if f not in self.Follow[B]:
                                    self.Follow[B].append(f)
                                    didadd = True
            if not didadd:
                break
        return self.Follow


    # -----------------------------------------------------------------------------
    # build_lritems()
    #
    # This function walks the list of productions and builds a complete set of the
    # LR items.  The LR items are stored in two ways:  First, they are uniquely
    # numbered and placed in the list _lritems.  Second, a linked list of LR items
    # is built for each production.  For example:
    #
    #   E -> E PLUS E
    #
    # Creates the list
    #
    #  [E -> . E PLUS E, E -> E . PLUS E, E -> E PLUS . E, E -> E PLUS E . ]
    # -----------------------------------------------------------------------------

    def build_lritems(self):
        for p in self.Productions:
            lastlri = p
            i = 0
            lr_items = []
            while True:
                if i > len(p):
                    lri = None
                else:
                    lri = LRItem(p, i)
                    # Precompute the list of productions immediately following
                    try:
                        lri.lr_after = self.Prodnames[lri.prod[i+1]]
                    except (IndexError, KeyError):
                        lri.lr_after = []
                    try:
                        lri.lr_before = lri.prod[i-1]
                    except IndexError:
                        lri.lr_before = None

                lastlri.lr_next = lri
                if not lri:
                    break
                lr_items.append(lri)
                lastlri = lri
                i += 1
            p.lr_items = lr_items


    # ----------------------------------------------------------------------
    # Debugging output.  Printing the grammar will produce a detailed
    # description along with some diagnostics.
    # ----------------------------------------------------------------------
    def __str__(self):
        out = []
        out.append('Grammar:\n')
        for n, p in enumerate(self.Productions):
            out.append(f'Rule {n:<5d} {p}')
        
        unused_terminals = self.unused_terminals()
        if unused_terminals:
            out.append('\nUnused terminals:\n')
            for term in unused_terminals:
                out.append(f'    {term}')

        out.append('\nTerminals, with rules where they appear:\n')
        for term in sorted(self.Terminals):
            out.append('%-20s : %s' % (term, ' '.join(str(s) for s in self.Terminals[term])))

        out.append('\nNonterminals, with rules where they appear:\n')
        for nonterm in sorted(self.Nonterminals):
            out.append('%-20s : %s' % (nonterm, ' '.join(str(s) for s in self.Nonterminals[nonterm])))

        out.append('')
        return '\n'.join(out)

# -----------------------------------------------------------------------------
#                           === LR Generator ===
#
# The following classes and functions are used to generate LR parsing tables on
# a grammar.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# digraph()
# traverse()
#
# The following two functions are used to compute set valued functions
# of the form:
#
#     F(x) = F'(x) U U{F(y) | x R y}
#
# This is used to compute the values of Read() sets as well as FOLLOW sets
# in LALR(1) generation.
#
# Inputs:  X    - An input set
#          R    - A relation
#          FP   - Set-valued function
# ------------------------------------------------------------------------------

def digraph(X, R, FP):
    N = {}
    for x in X:
        N[x] = 0
    stack = []
    F = {}
    for x in X:
        if N[x] == 0:
            traverse(x, N, stack, F, X, R, FP)
    return F

def traverse(x, N, stack, F, X, R, FP):
    stack.append(x)
    d = len(stack)
    N[x] = d
    F[x] = FP(x)             # F(X) <- F'(x)

    rel = R(x)               # Get y's related to x
    for y in rel:
        if N[y] == 0:
            traverse(y, N, stack, F, X, R, FP)
        N[x] = min(N[x], N[y])
        for a in F.get(y, []):
            if a not in F[x]:
                F[x].append(a)
    if N[x] == d:
        N[stack[-1]] = MAXINT
        F[stack[-1]] = F[x]
        element = stack.pop()
        while element != x:
            N[stack[-1]] = MAXINT
            F[stack[-1]] = F[x]
            element = stack.pop()

class LALRError(YaccError):
    pass

# -----------------------------------------------------------------------------
#                             == LRGeneratedTable ==
#
# This class implements the LR table generation algorithm.  There are no
# public methods except for write()
# -----------------------------------------------------------------------------

class LRTable(object):
    def __init__(self, grammar):
        self.grammar = grammar

        # Internal attributes
        self.lr_action     = {}        # Action table
        self.lr_goto       = {}        # Goto table
        self.lr_productions  = grammar.Productions    # Copy of grammar Production array
        self.lr_goto_cache = {}        # Cache of computed gotos
        self.lr0_cidhash   = {}        # Cache of closures
        self._add_count    = 0         # Internal counter used to detect cycles

        # Diagonistic information filled in by the table generator
        self.state_descriptions = OrderedDict()
        self.sr_conflict   = 0
        self.rr_conflict   = 0
        self.conflicts     = []        # List of conflicts

        self.sr_conflicts  = []
        self.rr_conflicts  = []

        # Build the tables
        self.grammar.build_lritems()
        self.grammar.compute_first()
        self.grammar.compute_follow()
        self.lr_parse_table()

        # Build default states
        # This identifies parser states where there is only one possible reduction action.
        # For such states, the parser can make a choose to make a rule reduction without consuming
        # the next look-ahead token.  This delayed invocation of the tokenizer can be useful in
        # certain kinds of advanced parsing situations where the lexer and parser interact with
        # each other or change states (i.e., manipulation of scope, lexer states, etc.).
        #
        # See:  http://www.gnu.org/software/bison/manual/html_node/Default-Reductions.html#Default-Reductions
        self.defaulted_states = {}
        for state, actions in self.lr_action.items():
            rules = list(actions.values())
            if len(rules) == 1 and rules[0] < 0:
                self.defaulted_states[state] = rules[0]

    # Compute the LR(0) closure operation on I, where I is a set of LR(0) items.
    def lr0_closure(self, I):
        self._add_count += 1

        # Add everything in I to J
        J = I[:]
        didadd = True
        while didadd:
            didadd = False
            for j in J:
                for x in j.lr_after:
                    if getattr(x, 'lr0_added', 0) == self._add_count:
                        continue
                    # Add B --> .G to J
                    J.append(x.lr_next)
                    x.lr0_added = self._add_count
                    didadd = True

        return J

    # Compute the LR(0) goto function goto(I,X) where I is a set
    # of LR(0) items and X is a grammar symbol.   This function is written
    # in a way that guarantees uniqueness of the generated goto sets
    # (i.e. the same goto set will never be returned as two different Python
    # objects).  With uniqueness, we can later do fast set comparisons using
    # id(obj) instead of element-wise comparison.

    def lr0_goto(self, I, x):
        # First we look for a previously cached entry
        g = self.lr_goto_cache.get((id(I), x))
        if g:
            return g

        # Now we generate the goto set in a way that guarantees uniqueness
        # of the result

        s = self.lr_goto_cache.get(x)
        if not s:
            s = {}
            self.lr_goto_cache[x] = s

        gs = []
        for p in I:
            n = p.lr_next
            if n and n.lr_before == x:
                s1 = s.get(id(n))
                if not s1:
                    s1 = {}
                    s[id(n)] = s1
                gs.append(n)
                s = s1
        g = s.get('$end')
        if not g:
            if gs:
                g = self.lr0_closure(gs)
                s['$end'] = g
            else:
                s['$end'] = gs
        self.lr_goto_cache[(id(I), x)] = g
        return g

    # Compute the LR(0) sets of item function
    def lr0_items(self):
        C = [self.lr0_closure([self.grammar.Productions[0].lr_next])]
        i = 0
        for I in C:
            self.lr0_cidhash[id(I)] = i
            i += 1

        # Loop over the items in C and each grammar symbols
        i = 0
        while i < len(C):
            I = C[i]
            i += 1

            # Collect all of the symbols that could possibly be in the goto(I,X) sets
            asyms = {}
            for ii in I:
                for s in ii.usyms:
                    asyms[s] = None

            for x in asyms:
                g = self.lr0_goto(I, x)
                if not g or id(g) in self.lr0_cidhash:
                    continue
                self.lr0_cidhash[id(g)] = len(C)
                C.append(g)

        return C

    # -----------------------------------------------------------------------------
    #                       ==== LALR(1) Parsing ====
    #
    # LALR(1) parsing is almost exactly the same as SLR except that instead of
    # relying upon Follow() sets when performing reductions, a more selective
    # lookahead set that incorporates the state of the LR(0) machine is utilized.
    # Thus, we mainly just have to focus on calculating the lookahead sets.
    #
    # The method used here is due to DeRemer and Pennelo (1982).
    #
    # DeRemer, F. L., and T. J. Pennelo: "Efficient Computation of LALR(1)
    #     Lookahead Sets", ACM Transactions on Programming Languages and Systems,
    #     Vol. 4, No. 4, Oct. 1982, pp. 615-649
    #
    # Further details can also be found in:
    #
    #  J. Tremblay and P. Sorenson, "The Theory and Practice of Compiler Writing",
    #      McGraw-Hill Book Company, (1985).
    #
    # -----------------------------------------------------------------------------

    # -----------------------------------------------------------------------------
    # compute_nullable_nonterminals()
    #
    # Creates a dictionary containing all of the non-terminals that might produce
    # an empty production.
    # -----------------------------------------------------------------------------

    def compute_nullable_nonterminals(self):
        nullable = set()
        num_nullable = 0
        while True:
            for p in self.grammar.Productions[1:]:
                if p.len == 0:
                    nullable.add(p.name)
                    continue
                for t in p.prod:
                    if t not in nullable:
                        break
                else:
                    nullable.add(p.name)
            if len(nullable) == num_nullable:
                break
            num_nullable = len(nullable)
        return nullable

    # -----------------------------------------------------------------------------
    # find_nonterminal_trans(C)
    #
    # Given a set of LR(0) items, this functions finds all of the non-terminal
    # transitions.    These are transitions in which a dot appears immediately before
    # a non-terminal.   Returns a list of tuples of the form (state,N) where state
    # is the state number and N is the nonterminal symbol.
    #
    # The input C is the set of LR(0) items.
    # -----------------------------------------------------------------------------

    def find_nonterminal_transitions(self, C):
        trans = []
        for stateno, state in enumerate(C):
            for p in state:
                if p.lr_index < p.len - 1:
                    t = (stateno, p.prod[p.lr_index+1])
                    if t[1] in self.grammar.Nonterminals:
                        if t not in trans:
                            trans.append(t)
        return trans

    # -----------------------------------------------------------------------------
    # dr_relation()
    #
    # Computes the DR(p,A) relationships for non-terminal transitions.  The input
    # is a tuple (state,N) where state is a number and N is a nonterminal symbol.
    #
    # Returns a list of terminals.
    # -----------------------------------------------------------------------------

    def dr_relation(self, C, trans, nullable):
        dr_set = {}
        state, N = trans
        terms = []

        g = self.lr0_goto(C[state], N)
        for p in g:
            if p.lr_index < p.len - 1:
                a = p.prod[p.lr_index+1]
                if a in self.grammar.Terminals:
                    if a not in terms:
                        terms.append(a)

        # This extra bit is to handle the start state
        if state == 0 and N == self.grammar.Productions[0].prod[0]:
            terms.append('$end')

        return terms

    # -----------------------------------------------------------------------------
    # reads_relation()
    #
    # Computes the READS() relation (p,A) READS (t,C).
    # -----------------------------------------------------------------------------

    def reads_relation(self, C, trans, empty):
        # Look for empty transitions
        rel = []
        state, N = trans

        g = self.lr0_goto(C[state], N)
        j = self.lr0_cidhash.get(id(g), -1)
        for p in g:
            if p.lr_index < p.len - 1:
                a = p.prod[p.lr_index + 1]
                if a in empty:
                    rel.append((j, a))

        return rel

    # -----------------------------------------------------------------------------
    # compute_lookback_includes()
    #
    # Determines the lookback and includes relations
    #
    # LOOKBACK:
    #
    # This relation is determined by running the LR(0) state machine forward.
    # For example, starting with a production "N : . A B C", we run it forward
    # to obtain "N : A B C ."   We then build a relationship between this final
    # state and the starting state.   These relationships are stored in a dictionary
    # lookdict.
    #
    # INCLUDES:
    #
    # Computes the INCLUDE() relation (p,A) INCLUDES (p',B).
    #
    # This relation is used to determine non-terminal transitions that occur
    # inside of other non-terminal transition states.   (p,A) INCLUDES (p', B)
    # if the following holds:
    #
    #       B -> LAT, where T -> epsilon and p' -L-> p
    #
    # L is essentially a prefix (which may be empty), T is a suffix that must be
    # able to derive an empty string.  State p' must lead to state p with the string L.
    #
    # -----------------------------------------------------------------------------

    def compute_lookback_includes(self, C, trans, nullable):
        lookdict = {}          # Dictionary of lookback relations
        includedict = {}       # Dictionary of include relations

        # Make a dictionary of non-terminal transitions
        dtrans = {}
        for t in trans:
            dtrans[t] = 1

        # Loop over all transitions and compute lookbacks and includes
        for state, N in trans:
            lookb = []
            includes = []
            for p in C[state]:
                if p.name != N:
                    continue

                # Okay, we have a name match.  We now follow the production all the way
                # through the state machine until we get the . on the right hand side

                lr_index = p.lr_index
                j = state
                while lr_index < p.len - 1:
                    lr_index = lr_index + 1
                    t = p.prod[lr_index]

                    # Check to see if this symbol and state are a non-terminal transition
                    if (j, t) in dtrans:
                        # Yes.  Okay, there is some chance that this is an includes relation
                        # the only way to know for certain is whether the rest of the
                        # production derives empty

                        li = lr_index + 1
                        while li < p.len:
                            if p.prod[li] in self.grammar.Terminals:
                                break      # No forget it
                            if p.prod[li] not in nullable:
                                break
                            li = li + 1
                        else:
                            # Appears to be a relation between (j,t) and (state,N)
                            includes.append((j, t))

                    g = self.lr0_goto(C[j], t)               # Go to next set
                    j = self.lr0_cidhash.get(id(g), -1)      # Go to next state

                # When we get here, j is the final state, now we have to locate the production
                for r in C[j]:
                    if r.name != p.name:
                        continue
                    if r.len != p.len:
                        continue
                    i = 0
                    # This look is comparing a production ". A B C" with "A B C ."
                    while i < r.lr_index:
                        if r.prod[i] != p.prod[i+1]:
                            break
                        i = i + 1
                    else:
                        lookb.append((j, r))
            for i in includes:
                if i not in includedict:
                    includedict[i] = []
                includedict[i].append((state, N))
            lookdict[(state, N)] = lookb

        return lookdict, includedict

    # -----------------------------------------------------------------------------
    # compute_read_sets()
    #
    # Given a set of LR(0) items, this function computes the read sets.
    #
    # Inputs:  C        =  Set of LR(0) items
    #          ntrans   = Set of nonterminal transitions
    #          nullable = Set of empty transitions
    #
    # Returns a set containing the read sets
    # -----------------------------------------------------------------------------

    def compute_read_sets(self, C, ntrans, nullable):
        FP = lambda x: self.dr_relation(C, x, nullable)
        R =  lambda x: self.reads_relation(C, x, nullable)
        F = digraph(ntrans, R, FP)
        return F

    # -----------------------------------------------------------------------------
    # compute_follow_sets()
    #
    # Given a set of LR(0) items, a set of non-terminal transitions, a readset,
    # and an include set, this function computes the follow sets
    #
    # Follow(p,A) = Read(p,A) U U {Follow(p',B) | (p,A) INCLUDES (p',B)}
    #
    # Inputs:
    #            ntrans     = Set of nonterminal transitions
    #            readsets   = Readset (previously computed)
    #            inclsets   = Include sets (previously computed)
    #
    # Returns a set containing the follow sets
    # -----------------------------------------------------------------------------

    def compute_follow_sets(self, ntrans, readsets, inclsets):
        FP = lambda x: readsets[x]
        R  = lambda x: inclsets.get(x, [])
        F = digraph(ntrans, R, FP)
        return F

    # -----------------------------------------------------------------------------
    # add_lookaheads()
    #
    # Attaches the lookahead symbols to grammar rules.
    #
    # Inputs:    lookbacks         -  Set of lookback relations
    #            followset         -  Computed follow set
    #
    # This function directly attaches the lookaheads to productions contained
    # in the lookbacks set
    # -----------------------------------------------------------------------------

    def add_lookaheads(self, lookbacks, followset):
        for trans, lb in lookbacks.items():
            # Loop over productions in lookback
            for state, p in lb:
                if state not in p.lookaheads:
                    p.lookaheads[state] = []
                f = followset.get(trans, [])
                for a in f:
                    if a not in p.lookaheads[state]:
                        p.lookaheads[state].append(a)

    # -----------------------------------------------------------------------------
    # add_lalr_lookaheads()
    #
    # This function does all of the work of adding lookahead information for use
    # with LALR parsing
    # -----------------------------------------------------------------------------

    def add_lalr_lookaheads(self, C):
        # Determine all of the nullable nonterminals
        nullable = self.compute_nullable_nonterminals()

        # Find all non-terminal transitions
        trans = self.find_nonterminal_transitions(C)

        # Compute read sets
        readsets = self.compute_read_sets(C, trans, nullable)

        # Compute lookback/includes relations
        lookd, included = self.compute_lookback_includes(C, trans, nullable)

        # Compute LALR FOLLOW sets
        followsets = self.compute_follow_sets(trans, readsets, included)

        # Add all of the lookaheads
        self.add_lookaheads(lookd, followsets)

    # -----------------------------------------------------------------------------
    # lr_parse_table()
    #
    # This function constructs the final LALR parse table.  Touch this code and die.
    # -----------------------------------------------------------------------------
    def lr_parse_table(self):
        Productions = self.grammar.Productions
        Precedence  = self.grammar.Precedence
        goto   = self.lr_goto         # Goto array
        action = self.lr_action       # Action array

        actionp = {}                  # Action production array (temporary)

        # Step 1: Construct C = { I0, I1, ... IN}, collection of LR(0) items
        # This determines the number of states

        C = self.lr0_items()
        self.add_lalr_lookaheads(C)

        # Build the parser table, state by state
        for st, I in enumerate(C):
            descrip = []
            # Loop over each production in I
            actlist = []              # List of actions
            st_action  = {}
            st_actionp = {}
            st_goto    = {}

            descrip.append(f'\nstate {st}\n')
            for p in I:
                descrip.append(f'    ({p.number}) {p}')

            for p in I:
                    if p.len == p.lr_index + 1:
                        if p.name == "S'":
                            # Start symbol. Accept!
                            st_action['$end'] = 0
                            st_actionp['$end'] = p
                        else:
                            # We are at the end of a production.  Reduce!
                            laheads = p.lookaheads[st]
                            for a in laheads:
                                actlist.append((a, p, f'reduce using rule {p.number} ({p})'))
                                r = st_action.get(a)
                                if r is not None:
                                    # Have a shift/reduce or reduce/reduce conflict
                                    if r > 0:
                                        # Need to decide on shift or reduce here
                                        # By default we favor shifting. Need to add
                                        # some precedence rules here.

                                        # Shift precedence comes from the token
                                        sprec, slevel = Precedence.get(a, ('right', 0))

                                        # Reduce precedence comes from rule being reduced (p)
                                        rprec, rlevel = Productions[p.number].prec

                                        if (slevel < rlevel) or ((slevel == rlevel) and (rprec == 'left')):
                                            # We really need to reduce here.
                                            st_action[a] = -p.number
                                            st_actionp[a] = p
                                            if not slevel and not rlevel:
                                                descrip.append(f'  ! shift/reduce conflict for {a} resolved as reduce')
                                                self.sr_conflicts.append((st, a, 'reduce'))
                                            Productions[p.number].reduced += 1
                                        elif (slevel == rlevel) and (rprec == 'nonassoc'):
                                            st_action[a] = None
                                        else:
                                            # Hmmm. Guess we'll keep the shift
                                            if not rlevel:
                                                descrip.append(f'  ! shift/reduce conflict for {a} resolved as shift')
                                                self.sr_conflicts.append((st, a, 'shift'))
                                    elif r <= 0:
                                        # Reduce/reduce conflict.   In this case, we favor the rule
                                        # that was defined first in the grammar file
                                        oldp = Productions[-r]
                                        pp = Productions[p.number]
                                        if oldp.line > pp.line:
                                            st_action[a] = -p.number
                                            st_actionp[a] = p
                                            chosenp, rejectp = pp, oldp
                                            Productions[p.number].reduced += 1
                                            Productions[oldp.number].reduced -= 1
                                        else:
                                            chosenp, rejectp = oldp, pp
                                        self.rr_conflicts.append((st, chosenp, rejectp))
                                        descrip.append('  ! reduce/reduce conflict for %s resolved using rule %d (%s)' % 
                                                       (a, st_actionp[a].number, st_actionp[a]))
                                    else:
                                        raise LALRError(f'Unknown conflict in state {st}')
                                else:
                                    st_action[a] = -p.number
                                    st_actionp[a] = p
                                    Productions[p.number].reduced += 1
                    else:
                        i = p.lr_index
                        a = p.prod[i+1]       # Get symbol right after the "."
                        if a in self.grammar.Terminals:
                            g = self.lr0_goto(I, a)
                            j = self.lr0_cidhash.get(id(g), -1)
                            if j >= 0:
                                # We are in a shift state
                                actlist.append((a, p, f'shift and go to state {j}'))
                                r = st_action.get(a)
                                if r is not None:
                                    # Whoa have a shift/reduce or shift/shift conflict
                                    if r > 0:
                                        if r != j:
                                            raise LALRError(f'Shift/shift conflict in state {st}')
                                    elif r <= 0:
                                        # Do a precedence check.
                                        #   -  if precedence of reduce rule is higher, we reduce.
                                        #   -  if precedence of reduce is same and left assoc, we reduce.
                                        #   -  otherwise we shift
                                        rprec, rlevel = Productions[st_actionp[a].number].prec
                                        sprec, slevel = Precedence.get(a, ('right', 0))
                                        if (slevel > rlevel) or ((slevel == rlevel) and (rprec == 'right')):
                                            # We decide to shift here... highest precedence to shift
                                            Productions[st_actionp[a].number].reduced -= 1
                                            st_action[a] = j
                                            st_actionp[a] = p
                                            if not rlevel:
                                                descrip.append(f'  ! shift/reduce conflict for {a} resolved as shift')
                                                self.sr_conflicts.append((st, a, 'shift'))
                                        elif (slevel == rlevel) and (rprec == 'nonassoc'):
                                            st_action[a] = None
                                        else:
                                            # Hmmm. Guess we'll keep the reduce
                                            if not slevel and not rlevel:
                                                descrip.append(f'  ! shift/reduce conflict for {a} resolved as reduce')
                                                self.sr_conflicts.append((st, a, 'reduce'))

                                    else:
                                        raise LALRError(f'Unknown conflict in state {st}')
                                else:
                                    st_action[a] = j
                                    st_actionp[a] = p

            # Print the actions associated with each terminal
            _actprint = {}
            for a, p, m in actlist:
                if a in st_action:
                    if p is st_actionp[a]:
                        descrip.append(f'    {a:<15s} {m}')
                        _actprint[(a, m)] = 1
            descrip.append('')

            # Construct the goto table for this state
            nkeys = {}
            for ii in I:
                for s in ii.usyms:
                    if s in self.grammar.Nonterminals:
                        nkeys[s] = None
            for n in nkeys:
                g = self.lr0_goto(I, n)
                j = self.lr0_cidhash.get(id(g), -1)
                if j >= 0:
                    st_goto[n] = j
                    descrip.append(f'    {n:<30s} shift and go to state {j}')

            action[st] = st_action
            actionp[st] = st_actionp
            goto[st] = st_goto
            self.state_descriptions[st] = '\n'.join(descrip)

    # ----------------------------------------------------------------------
    # Debugging output.   Printing the LRTable object will produce a listing
    # of all of the states, conflicts, and other details.
    # ----------------------------------------------------------------------
    def __str__(self):
        out = []
        for descrip in self.state_descriptions.values():
            out.append(descrip)
            
        if self.sr_conflicts or self.rr_conflicts:
            out.append('\nConflicts:\n')

            for state, tok, resolution in self.sr_conflicts:
                out.append(f'shift/reduce conflict for {tok} in state {state} resolved as {resolution}')

            already_reported = set()
            for state, rule, rejected in self.rr_conflicts:
                if (state, id(rule), id(rejected)) in already_reported:
                    continue
                out.append(f'reduce/reduce conflict in state {state} resolved using rule {rule}')
                out.append(f'rejected rule ({rejected}) in state {state}')
                already_reported.add((state, id(rule), id(rejected)))

            warned_never = set()
            for state, rule, rejected in self.rr_conflicts:
                if not rejected.reduced and (rejected not in warned_never):
                    out.append(f'Rule ({rejected}) is never reduced')
                    warned_never.add(rejected)

        return '\n'.join(out)

# Collect grammar rules from a function
def _collect_grammar_rules(func):
    grammar = []
    while func:
        prodname = func.__name__
        unwrapped = inspect.unwrap(func)
        filename = unwrapped.__code__.co_filename
        lineno = unwrapped.__code__.co_firstlineno
        for rule, lineno in zip(func.rules, range(lineno+len(func.rules)-1, 0, -1)):
            syms = rule.split()
            ebnf_prod = []
            while ('{' in syms) or ('[' in syms):
                for s in syms:
                    if s == '[':
                        syms, prod = _replace_ebnf_optional(syms)
                        ebnf_prod.extend(prod)
                        break
                    elif s == '{':
                        syms, prod = _replace_ebnf_repeat(syms)
                        ebnf_prod.extend(prod)
                        break
                    elif '|' in s:
                        syms, prod = _replace_ebnf_choice(syms)
                        ebnf_prod.extend(prod)
                        break

            if syms[1:2] == [':'] or syms[1:2] == ['::=']:
                grammar.append((func, filename, lineno, syms[0], syms[2:]))
            else:
                grammar.append((func, filename, lineno, prodname, syms))
            grammar.extend(ebnf_prod)
            
        func = getattr(func, 'next_func', None)

    return grammar

# Replace EBNF repetition
def _replace_ebnf_repeat(syms):
    syms = list(syms)
    first = syms.index('{')
    end = syms.index('}', first)

    # Look for choices inside
    repeated_syms = syms[first+1:end]
    if any('|' in sym for sym in repeated_syms):
        repeated_syms, prods = _replace_ebnf_choice(repeated_syms)
    else:
        prods = []

    symname, moreprods = _generate_repeat_rules(repeated_syms)
    syms[first:end+1] = [symname]
    return syms, prods + moreprods

def _replace_ebnf_optional(syms):
    syms = list(syms)
    first = syms.index('[')
    end = syms.index(']', first)
    symname, prods = _generate_optional_rules(syms[first+1:end])
    syms[first:end+1] = [symname]
    return syms, prods

def _replace_ebnf_choice(syms):
    syms = list(syms)
    newprods = [ ]
    n = 0
    while n < len(syms):
        if '|' in syms[n]:
            symname, prods = _generate_choice_rules(syms[n].split('|'))
            syms[n] = symname
            newprods.extend(prods)
        n += 1
    return syms, newprods
    
# Generate grammar rules for repeated items
_gencount = 0

# Dictionary mapping name aliases generated by EBNF rules.

_name_aliases = { }

def _sanitize_symbols(symbols):
    for sym in symbols:
        if sym.startswith("'"):
            yield str(hex(ord(sym[1])))
        elif sym.isidentifier():
            yield sym
        else:
            yield sym.encode('utf-8').hex()
            
def _generate_repeat_rules(symbols):
    '''
    Symbols is a list of grammar symbols [ symbols ]. This
    generates code corresponding to these grammar construction:
  
       @('repeat : many')
       def repeat(self, p):
           return p.many

       @('repeat :')
       def repeat(self, p):
           return []

       @('many : many symbols')
       def many(self, p):
           p.many.append(symbols)
           return p.many

       @('many : symbols')
       def many(self, p):
           return [ p.symbols ]
    '''
    global _gencount
    _gencount += 1
    basename = f'_{_gencount}_' + '_'.join(_sanitize_symbols(symbols))
    name = f'{basename}_repeat'
    oname = f'{basename}_items'
    iname = f'{basename}_item'
    symtext = ' '.join(symbols)

    _name_aliases[name] = symbols

    productions = [ ]
    _ = _decorator

    @_(f'{name} : {oname}')
    def repeat(self, p):
        return getattr(p, oname)

    @_(f'{name} : ')
    def repeat2(self, p):
        return []
    productions.extend(_collect_grammar_rules(repeat))
    productions.extend(_collect_grammar_rules(repeat2))

    @_(f'{oname} : {oname} {iname}')
    def many(self, p):
        items = getattr(p, oname)
        items.append(getattr(p, iname))
        return items

    @_(f'{oname} : {iname}')
    def many2(self, p):
        return [ getattr(p, iname) ]

    productions.extend(_collect_grammar_rules(many))
    productions.extend(_collect_grammar_rules(many2))

    @_(f'{iname} : {symtext}')
    def item(self, p):
        return tuple(p)

    productions.extend(_collect_grammar_rules(item))
    return name, productions

def _generate_optional_rules(symbols):
    '''
    Symbols is a list of grammar symbols [ symbols ]. This
    generates code corresponding to these grammar construction:
  
       @('optional : symbols')
       def optional(self, p):
           return p.symbols

       @('optional :')
       def optional(self, p):
           return None
    '''
    global _gencount
    _gencount += 1
    basename = f'_{_gencount}_' + '_'.join(_sanitize_symbols(symbols))
    name = f'{basename}_optional'
    symtext = ' '.join(symbols)
    
    _name_aliases[name] = symbols

    productions = [ ]
    _ = _decorator

    no_values = (None,) * len(symbols)

    @_(f'{name} : {symtext}')
    def optional(self, p):
        return tuple(p)

    @_(f'{name} : ')
    def optional2(self, p):
        return no_values

    productions.extend(_collect_grammar_rules(optional))
    productions.extend(_collect_grammar_rules(optional2))
    return name, productions

def _generate_choice_rules(symbols):
    '''
    Symbols is a list of grammar symbols such as [ 'PLUS', 'MINUS' ].
    This generates code corresponding to the following construction:
    
    @('PLUS', 'MINUS')
    def choice(self, p):
        return p[0]
    '''
    global _gencount
    _gencount += 1
    basename = f'_{_gencount}_' + '_'.join(_sanitize_symbols(symbols))
    name = f'{basename}_choice'

    _ = _decorator
    productions = [ ]


    def choice(self, p):
        return p[0]
    choice.__name__ = name
    choice = _(*symbols)(choice)
    productions.extend(_collect_grammar_rules(choice))
    return name, productions
    
class ParserMetaDict(dict):
    '''
    Dictionary that allows decorated grammar rule functions to be overloaded
    '''
    def __setitem__(self, key, value):
        if key in self and callable(value) and hasattr(value, 'rules'):
            value.next_func = self[key]
            if not hasattr(value.next_func, 'rules'):
                raise GrammarError(f'Redefinition of {key}. Perhaps an earlier {key} is missing @_')
        super().__setitem__(key, value)
    
    def __getitem__(self, key):
        if key not in self and key.isupper() and key[:1] != '_':
            return key.upper()
        else:
            return super().__getitem__(key)

def _decorator(rule, *extra):
     rules = [rule, *extra]
     def decorate(func):
         func.rules = [ *getattr(func, 'rules', []), *rules[::-1] ]
         return func
     return decorate

class ParserMeta(type):
    @classmethod
    def __prepare__(meta, *args, **kwargs):
        d = ParserMetaDict()
        d['_'] = _decorator
        return d

    def __new__(meta, clsname, bases, attributes):
        del attributes['_']
        cls = super().__new__(meta, clsname, bases, attributes)
        cls._build(list(attributes.items()))
        return cls

class Parser(metaclass=ParserMeta):
    # Automatic tracking of position information
    track_positions = True
    
    # Logging object where debugging/diagnostic messages are sent
    log = SlyLogger(sys.stderr)     

    # Debugging filename where parsetab.out data can be written
    debugfile = None

    @classmethod
    def __validate_tokens(cls):
        if not hasattr(cls, 'tokens'):
            cls.log.error('No token list is defined')
            return False

        if not cls.tokens:
            cls.log.error('tokens is empty')
            return False

        if 'error' in cls.tokens:
            cls.log.error("Illegal token name 'error'. Is a reserved word")
            return False

        return True

    @classmethod
    def __validate_precedence(cls):
        if not hasattr(cls, 'precedence'):
            cls.__preclist = []
            return True

        preclist = []
        if not isinstance(cls.precedence, (list, tuple)):
            cls.log.error('precedence must be a list or tuple')
            return False

        for level, p in enumerate(cls.precedence, start=1):
            if not isinstance(p, (list, tuple)):
                cls.log.error(f'Bad precedence table entry {p!r}. Must be a list or tuple')
                return False

            if len(p) < 2:
                cls.log.error(f'Malformed precedence entry {p!r}. Must be (assoc, term, ..., term)')
                return False

            if not all(isinstance(term, str) for term in p):
                cls.log.error('precedence items must be strings')
                return False
            
            assoc = p[0]
            preclist.extend((term, assoc, level) for term in p[1:])

        cls.__preclist = preclist
        return True

    @classmethod
    def __validate_specification(cls):
        '''
        Validate various parts of the grammar specification
        '''
        if not cls.__validate_tokens():
            return False
        if not cls.__validate_precedence():
            return False
        return True

    @classmethod
    def __build_grammar(cls, rules):
        '''
        Build the grammar from the grammar rules
        '''
        grammar_rules = []
        errors = ''
        # Check for non-empty symbols
        if not rules:
            raise YaccError('No grammar rules are defined')

        grammar = Grammar(cls.tokens)

        # Set the precedence level for terminals
        for term, assoc, level in cls.__preclist:
            try:
                grammar.set_precedence(term, assoc, level)
            except GrammarError as e:
                errors += f'{e}\n'

        for name, func in rules:
            try:
                parsed_rule = _collect_grammar_rules(func)
                for pfunc, rulefile, ruleline, prodname, syms in parsed_rule:
                    try:
                        grammar.add_production(prodname, syms, pfunc, rulefile, ruleline)
                    except GrammarError as e:
                        errors += f'{e}\n'
            except SyntaxError as e:
                errors += f'{e}\n'
        try:
            grammar.set_start(getattr(cls, 'start', None))
        except GrammarError as e:
            errors += f'{e}\n'

        undefined_symbols = grammar.undefined_symbols()
        for sym, prod in undefined_symbols:
            errors += '%s:%d: Symbol %r used, but not defined as a token or a rule\n' % (prod.file, prod.line, sym)

        unused_terminals = grammar.unused_terminals()
        if unused_terminals:
            unused_str = '{' + ','.join(unused_terminals) + '}'
            cls.log.warning(f'Token{"(s)" if len(unused_terminals) >1 else ""} {unused_str} defined, but not used')

        unused_rules = grammar.unused_rules()
        for prod in unused_rules:
            cls.log.warning('%s:%d: Rule %r defined, but not used', prod.file, prod.line, prod.name)

        if len(unused_terminals) == 1:
            cls.log.warning('There is 1 unused token')
        if len(unused_terminals) > 1:
            cls.log.warning('There are %d unused tokens', len(unused_terminals))

        if len(unused_rules) == 1:
            cls.log.warning('There is 1 unused rule')
        if len(unused_rules) > 1:
            cls.log.warning('There are %d unused rules', len(unused_rules))

        unreachable = grammar.find_unreachable()
        for u in unreachable:
           cls.log.warning('Symbol %r is unreachable', u)

        if len(undefined_symbols) == 0:
            infinite = grammar.infinite_cycles()
            for inf in infinite:
                errors += 'Infinite recursion detected for symbol %r\n' % inf

        unused_prec = grammar.unused_precedence()
        for term, assoc in unused_prec:
            errors += 'Precedence rule %r defined for unknown symbol %r\n' % (assoc, term)

        cls._grammar = grammar
        if errors:
            raise YaccError('Unable to build grammar.\n'+errors)

    @classmethod
    def __build_lrtables(cls):
        '''
        Build the LR Parsing tables from the grammar
        '''
        lrtable = LRTable(cls._grammar)
        num_sr = len(lrtable.sr_conflicts)

        # Report shift/reduce and reduce/reduce conflicts
        if num_sr != getattr(cls, 'expected_shift_reduce', None):
            if num_sr == 1:
                cls.log.warning('1 shift/reduce conflict')
            elif num_sr > 1:
                cls.log.warning('%d shift/reduce conflicts', num_sr)

        num_rr = len(lrtable.rr_conflicts)
        if num_rr != getattr(cls, 'expected_reduce_reduce', None):
            if num_rr == 1:
                cls.log.warning('1 reduce/reduce conflict')
            elif num_rr > 1:
                cls.log.warning('%d reduce/reduce conflicts', num_rr)

        cls._lrtable = lrtable
        return True

    @classmethod
    def __collect_rules(cls, definitions):
        '''
        Collect all of the tagged grammar rules
        '''
        rules = [ (name, value) for name, value in definitions
                  if callable(value) and hasattr(value, 'rules') ]
        return rules

    # ----------------------------------------------------------------------
    # Build the LALR(1) tables. definitions is a list of (name, item) tuples
    # of all definitions provided in the class, listed in the order in which
    # they were defined.  This method is triggered by a metaclass.
    # ----------------------------------------------------------------------
    @classmethod
    def _build(cls, definitions):
        if vars(cls).get('_build', False):
            return

        # Collect all of the grammar rules from the class definition
        rules = cls.__collect_rules(definitions)

        # Validate other parts of the grammar specification
        if not cls.__validate_specification():
            raise YaccError('Invalid parser specification')

        # Build the underlying grammar object
        cls.__build_grammar(rules)

        # Build the LR tables
        if not cls.__build_lrtables():
            raise YaccError('Can\'t build parsing tables')

        if cls.debugfile:
            with open(cls.debugfile, 'w') as f:
                f.write(str(cls._grammar))
                f.write('\n')
                f.write(str(cls._lrtable))
            cls.log.info('Parser debugging for %s written to %s', cls.__qualname__, cls.debugfile)

    # ----------------------------------------------------------------------
    # Parsing Support.  This is the parsing runtime that users use to
    # ----------------------------------------------------------------------
    def error(self, token):
        '''
        Default error handling function.  This may be subclassed.
        '''
        if token:
            lineno = getattr(token, 'lineno', 0)
            if lineno:
                sys.stderr.write(f'sly: Syntax error at line {lineno}, token={token.type}\n')
            else:
                sys.stderr.write(f'sly: Syntax error, token={token.type}')
        else:
            sys.stderr.write('sly: Parse error in input. EOF\n')
 
    def errok(self):
        '''
        Clear the error status
        '''
        self.errorok = True

    def restart(self):
        '''
        Force the parser to restart from a fresh state. Clears the statestack
        '''
        del self.statestack[:]
        del self.symstack[:]
        sym = YaccSymbol()
        sym.type = '$end'
        self.symstack.append(sym)
        self.statestack.append(0)
        self.state = 0

    def parse(self, tokens):
        '''
        Parse the given input tokens.
        '''
        lookahead = None                                  # Current lookahead symbol
        lookaheadstack = []                               # Stack of lookahead symbols
        actions = self._lrtable.lr_action                 # Local reference to action table (to avoid lookup on self.)
        goto    = self._lrtable.lr_goto                   # Local reference to goto table (to avoid lookup on self.)
        prod    = self._grammar.Productions               # Local reference to production list (to avoid lookup on self.)
        defaulted_states = self._lrtable.defaulted_states # Local reference to defaulted states
        pslice  = YaccProduction(None)                    # Production object passed to grammar rules
        errorcount = 0                                    # Used during error recovery

        # Set up the state and symbol stacks
        self.tokens = tokens
        self.statestack = statestack = []                 # Stack of parsing states
        self.symstack = symstack = []                     # Stack of grammar symbols
        pslice._stack = symstack                          # Associate the stack with the production
        self.restart()

        # Set up position tracking
        track_positions = self.track_positions
        if not hasattr(self, '_line_positions'):
            self._line_positions = { }           # id: -> lineno
            self._index_positions = { }          # id: -> (start, end)

        errtoken   = None                                 # Err token
        while True:
            # Get the next symbol on the input.  If a lookahead symbol
            # is already set, we just use that. Otherwise, we'll pull
            # the next token off of the lookaheadstack or from the lexer
            if self.state not in defaulted_states:
                if not lookahead:
                    if not lookaheadstack:
                        lookahead = next(tokens, None)  # Get the next token
                    else:
                        lookahead = lookaheadstack.pop()
                    if not lookahead:
                        lookahead = YaccSymbol()
                        lookahead.type = '$end'
                    
                # Check the action table
                ltype = lookahead.type
                t = actions[self.state].get(ltype)
            else:
                t = defaulted_states[self.state]

            if t is not None:
                if t > 0:
                    # shift a symbol on the stack
                    statestack.append(t)
                    self.state = t

                    symstack.append(lookahead)
                    lookahead = None

                    # Decrease error count on successful shift
                    if errorcount:
                        errorcount -= 1
                    continue

                if t < 0:
                    # reduce a symbol on the stack, emit a production
                    self.production = p = prod[-t]
                    pname = p.name
                    plen  = p.len
                    pslice._namemap = p.namemap

                    # Call the production function
                    pslice._slice = symstack[-plen:] if plen else []

                    sym = YaccSymbol()
                    sym.type = pname       
                    value = p.func(self, pslice)
                    if value is pslice:
                        value = (pname, *(s.value for s in pslice._slice))

                    sym.value = value
                        
                    # Record positions
                    if track_positions:
                        if plen:
                            sym.lineno = symstack[-plen].lineno
                            sym.index = symstack[-plen].index
                            sym.end = symstack[-1].end
                        else:
                            # A zero-length production  (what to put here?)
                            sym.lineno = None
                            sym.index = None
                            sym.end = None
                        self._line_positions[id(value)] = sym.lineno
                        self._index_positions[id(value)] = (sym.index, sym.end)
                            
                    if plen:
                        del symstack[-plen:]
                        del statestack[-plen:]

                    symstack.append(sym)
                    self.state = goto[statestack[-1]][pname]
                    statestack.append(self.state)
                    continue

                if t == 0:
                    n = symstack[-1]
                    result = getattr(n, 'value', None)
                    return result

            if t is None:
                # We have some kind of parsing error here.  To handle
                # this, we are going to push the current token onto
                # the tokenstack and replace it with an 'error' token.
                # If there are any synchronization rules, they may
                # catch it.
                #
                # In addition to pushing the error token, we call call
                # the user defined error() function if this is the
                # first syntax error.  This function is only called if
                # errorcount == 0.
                if errorcount == 0 or self.errorok:
                    errorcount = ERROR_COUNT
                    self.errorok = False
                    if lookahead.type == '$end':
                        errtoken = None               # End of file!
                    else:
                        errtoken = lookahead

                    tok = self.error(errtoken)
                    if tok:
                        # User must have done some kind of panic
                        # mode recovery on their own.  The
                        # returned token is the next lookahead
                        lookahead = tok
                        self.errorok = True
                        continue
                    else:
                        # If at EOF. We just return. Basically dead.
                        if not errtoken:
                            return
                else:
                    # Reset the error count.  Unsuccessful token shifted
                    errorcount = ERROR_COUNT

                # case 1:  the statestack only has 1 entry on it.  If we're in this state, the
                # entire parse has been rolled back and we're completely hosed.   The token is
                # discarded and we just keep going.

                if len(statestack) <= 1 and lookahead.type != '$end':
                    lookahead = None
                    self.state = 0
                    # Nuke the lookahead stack
                    del lookaheadstack[:]
                    continue

                # case 2: the statestack has a couple of entries on it, but we're
                # at the end of the file. nuke the top entry and generate an error token

                # Start nuking entries on the stack
                if lookahead.type == '$end':
                    # Whoa. We're really hosed here. Bail out
                    return

                if lookahead.type != 'error':
                    sym = symstack[-1]
                    if sym.type == 'error':
                        # Hmmm. Error is on top of stack, we'll just nuke input
                        # symbol and continue
                        lookahead = None
                        continue

                    # Create the error symbol for the first time and make it the new lookahead symbol
                    t = YaccSymbol()
                    t.type = 'error'

                    if hasattr(lookahead, 'lineno'):
                        t.lineno = lookahead.lineno
                    if hasattr(lookahead, 'index'):
                        t.index = lookahead.index
                    if hasattr(lookahead, 'end'):
                        t.end = lookahead.end
                    t.value = lookahead
                    lookaheadstack.append(lookahead)
                    lookahead = t
                else:
                    sym = symstack.pop()
                    statestack.pop()
                    self.state = statestack[-1]
                continue

            # Call an error function here
            raise RuntimeError('sly: internal parser error!!!\n')

    # Return position tracking information
    def line_position(self, value):
        return self._line_positions[id(value)]

    def index_position(self, value):
        return self._index_positions[id(value)]
    



"""
This script provides the abstract SqlCreator and concrete classes for Insert, Update, and Delete SQL creators.

Turning a diff into SQL statements works as follows:

1. Iterate rows in diff, because each query depends on the values in a row. Then repeat the following steps for each row.
2. Create a dictionary with unique column headers as keys and values as values. We'll need these for all query types.
3. Create a dictionary with other column headers as keys and values as values. Include non-null values for inserts and modified values for updates. Leave it empty for deletes.
4. Filter tree based on keys in the two dictionaries.
5. Compile the filtered tree to get parameter names for the query.
6. For each column in the tree, look up values by header using the dictionaries created earlier
7. Split values if the column has more than one parameter names. A CSV cell can contain multiple values separated by a colon
7. Clean up values. Strip whitespace from strings. Convert values to match the DB schema. For example, convert '1' to 1 if the column is an int
8. Create the dictionary with parameter names and values.
9. Compile the filtered tree to get the query for a row. Use the parameter dictionary where needed. For example, in SQL you write 'x = 1' but 'x is null'
10. Return query and parameter dictionary

Author: Romke Jonker
Email: romke@rnadesign.net
"""
from abc import ABC, abstractmethod

import numpy
import pandas as pd
from psycopg2._json import Json

from .query_executor import FailedQueryExecutor
from ..stml.header_renderer import HeaderRenderer
from ..stml.model import Entity
from ..stml.sql.parameter_types_renderer import ParameterTypesRenderer
from ..stml.sql.parameters_renderer import ParametersRenderer
from ..stml.values_parser import ValuesLexer, ValuesParser


class ExecutorCreator(ABC):

    def __init__(self):
        self._values_lexer = ValuesLexer()
        self._values_parser = ValuesParser()
        self.operation_type = None

    def create_executors(self, mapping, diffs, context=None, orm=None):

        # iterate rows in diff
        for i in range(len(diffs)):
            row = diffs.iloc[i]

            # return line number, depends on operation type
            line_number = self._get_line_number(row)

            try:
                # create and yield executor
                yield self._prepare_and_create_executor(mapping, row, line_number, context, orm)

            except Exception as e:

                # yield query with line number and error message
                yield FailedQueryExecutor(line_number, self.operation_type, mapping.name, context, str(e))

    def _prepare_and_create_executor(self, mapping, row, line_number, context=None, orm=None):
        # prepare mapping and values for row
        filtered_mapping, value_dict = self._prepare_mapping_and_values(mapping, row)

        # create and return executor
        return self._create_executor(line_number, filtered_mapping, value_dict, context, orm)

    def _prepare_mapping_and_values(self, mapping, row):
        # Create a dictionary with unique column headers as keys and values as values. We'll need these for all query types.
        unique_value_dict = self._create_unique_value_dict(mapping, row)

        # raise error if there are no unique values
        assert unique_value_dict, 'Header must have at least one unique column'

        # Create a dictionary with other column headers as keys and values as values.
        non_unique_value_dict = self._create_non_unique_value_dict(mapping, row)

        # create combined dictionary for convenience
        header_value_dict = {**unique_value_dict, **non_unique_value_dict}

        # Filter tree based on keys in the combined dictionary
        filtered_mapping = self._filter_mapping(mapping, header_value_dict)

        # Compile the filtered tree to get parameter names for the query.
        parameter_names = ParametersRenderer().render(filtered_mapping)

        # Create the dictionary with parameter names as keys and values as values. An element may contain multiple parameters and values
        parameter_value_dict = self._map_parameter_names_with_values(filtered_mapping, parameter_names, header_value_dict)

        # Split values if the column has more than one parameter names. A CSV cell can contain multiple values separated by a colon
        split_parameter_value_dict = self._split_columns(parameter_names, parameter_value_dict)

        # get parameter types
        parameter_types = ParameterTypesRenderer().render(filtered_mapping)

        # Clean up values. Strip whitespace from strings. Convert values to match the DB schema. For example, convert '1' to 1 if the column is an int
        value_dict_clean = self._clean_values_in_dict(split_parameter_value_dict, parameter_types)

        return filtered_mapping, value_dict_clean

    def _create_unique_value_dict(self, mapping, row):
        # get unique column headers
        unique_headers = HeaderRenderer().render_list_unique(mapping)

        # create dictionary with unique column headers as keys and values as values
        unique_value_dict = {header: row[header] for header in unique_headers}

        return unique_value_dict

    def _create_non_unique_value_dict(self, mapping, row):
        # get non-unique headers
        non_unique_headers = HeaderRenderer().render_list_non_unique(mapping)

        # create dictionary with non-unique column headers as keys and values as values
        non_unique_value_dict = {header: row[header] for header in non_unique_headers}

        return non_unique_value_dict

    def _create_non_unique_root_extension_value_dict(self, mapping, row):
        # when we delete a record with an extension on the root table (not on a foreign key table),
        # then we need to delete the extension record as well. If the extension is unique, then
        # the delete statement will already have the extension's name value, but if it's not unique,
        # then this method ensures the value is included.

        # get non-unique headers
        non_unique_headers = HeaderRenderer().render_list_non_unique_root_extension(mapping)

        # create dictionary with non-unique column headers as keys and values as values
        non_unique_value_dict = {header: row[header] for header in non_unique_headers}

        return non_unique_value_dict

    def _filter_mapping(self, mapping: Entity, value_dict):
        # filter columns by those that have a value in value_dict

        # create a list of column headers
        headers = HeaderRenderer().render_list(mapping, include_skip=True, include_orm_only=True)

        # create a list of all columns
        all_columns = mapping.attributes

        assert len(headers) == len(all_columns), f'Number of headers must equal number of columns, found: {len(headers)} and {len(all_columns)}'

        # create a list of items in mapping where the corresponding header is in value_dict
        filter_columns = [item for item, header in zip(all_columns, headers) if header in value_dict]

        # create a copy of mapping with columns replaced by filtered columns
        filtered_mapping = Entity(mapping.name)
        filtered_mapping.attributes = filter_columns

        return filtered_mapping

    def _map_parameter_names_with_values(self, filtered_mapping, parameter_names, value_dict):
        # create a list of column headers. Include the ORM-only columns, because we need them to write to ORM.
        headers = HeaderRenderer().render_list(filtered_mapping, include_orm_only=True)

        # create a dictionary with parameter names as keys and values as values
        return {parameter_name: value_dict[header] for parameter_name, header in zip(parameter_names, headers)}

    def _split_columns(self, parameter_names, value_dict):
        # verify that parameter_names and value_dict have the same number of items
        assert len(parameter_names) == len(value_dict), f'Number of parameter names must equal number of values, found: {parameter_names} and {value_dict}'

        # create empty dictionary
        value_dict_split = {}

        # iterate parameter names
        for parameter_name in parameter_names:
            # split the value and add it to the dictionary
            value_dict_split.update(self._split_value(parameter_name, value_dict[parameter_name]))

        return value_dict_split

    def _split_value(self, names, values):
        # if there is only one name, there is only one value
        if len(names) == 1:
            return {names[0]: values}

        # else split values at colon, but make sure not to split at json colons
        split_values = self._values_parser.parse(self._values_lexer.tokenize(values))

        # verify that number of names equals number of values
        assert len(names) == len(split_values), f'Number of names must equal number of values, found: {names} and {split_values}'

        # create dictionary with parameter names as keys and values as values
        return {name: value for name, value in zip(names, split_values)}

    def _clean_values_in_dict(self, value_dict, parameter_types):
        return {key: self._clean_value_in_dict(value, parameter_types[key]) for key, value in value_dict.items()}

    def _clean_value_in_dict(self, value, type):
        # strip whitespace from strings only
        if isinstance(value, str):
            value = value.strip()

        # convert numpy.int64 to int
        if isinstance(value, numpy.int64):
            value = int(value)

        # convert numpy.float64 to float
        if isinstance(value, numpy.float64):
            value = float(value)

        # convert numpy bool to python bool
        if isinstance(value, numpy.bool_):
            # psycopg wants string representations for bools
            value = 'true' if value else 'false'

        # convert dict to json string
        if isinstance(value, dict):
            # use psycopg's Json to correctly convert unicode characters
            value = Json(value)

        # dict is stored in data frame as frozenset, bec/ the index requires hashable objects
        if isinstance(value, frozenset):
            # convert to dict, then use psycopg's Json to correctly convert unicode characters
            value = Json(dict(value))

        # type conversions
        if type == 'int':
            value = int(value)
        elif type in ['varchar', 'text']:
            value = str(value)
        elif type == 'float':
            value = float(value)
        elif type == 'bool':
            # psycopg wants string representations for bools
            value = 'true' if value else 'false'

        return value

    def _is_empty(self, value):
        return value is None or pd.isnull(value) or pd.isna(value) or value == ''

    @abstractmethod
    def _create_executor(self, line_number, mapping, value_dict, context, orm):
        raise NotImplementedError

    def _get_line_number(self, row):
        # for inserts and updates, just return __line__ column
        return row['__line__']



"""
This class adds alias names to tables and parameter names to terminal (not foreign key) columns.
The reason we shouldn't create parameter names on the fly when generating SQL queries, is that the order in
which the sql compilers process cells is hard to predict. For example, UpdateCompiler leaves parsing
of unique columns to the end (when generating the where clause). This makes it particularly hard to
create column headers with the correct parameter names.
This compiler generates parameter names in a single linear parsing round.

Author: Romke Jonker
Email: romke@stml.io
"""

from stimula.stml.model import Entity, AbstractAttribute, Attribute, Reference


class AliasEnricher:
    def __init__(self):
        self._table_suffixes = {}
        self._parameter_suffixes = {}

    def enrich(self, entity: Entity):
        # must not re-use this instance
        assert not self._table_suffixes, 'AliasEnricher instance must not be re-used'

        # initialize table suffix at 0
        self._table_suffixes[entity.name] = 0

        # iterate all properties
        [self._property(p) for p in entity.attributes]
        return entity

    def _property(self, property: AbstractAttribute):
        if isinstance(property, Attribute):
            # set unique parameter name so queries use predictable names
            property.parameter = self._get_unique_parameter_id(property.name)

        if isinstance(property, Reference):
            # set alias
            property.alias = self._get_unique_table_id(property.table)

            # recurse into properties
            [self._property(p) for p in property.attributes]


    def _get_unique_table_id(self, name):
        if name not in self._table_suffixes:
            self._table_suffixes[name] = 0
            return name
        # increase suffix to make unique
        self._table_suffixes[name] += 1
        return name + '_' + str(self._table_suffixes[name])

    def _get_unique_parameter_id(self, name):
        if name not in self._parameter_suffixes:
            self._parameter_suffixes[name] = 0
            return name
        # increase suffix to make unique
        self._parameter_suffixes[name] += 1
        return name + '_' + str(self._parameter_suffixes[name])



'''
Compiles foreign key where clauses for insert and update.
This compiler is used to filter the foreign key tables when inserting or updating records to obtain the foreign key values.
'''

from itertools import chain
from typing import List

from stimula.stml.model import Entity, AbstractAttribute, Attribute, Reference


class ForeignWhereClauseRenderer:

    def __init__(self, is_insert_query, is_delete_query):
        # if this is an insert query, we must not join with extension tables
        self._is_insert_query = is_insert_query
        self._is_delete_query = is_delete_query
        self._aliases = {}

    def render(self, mapping: Entity) -> str:
        # no need to list non-unique columns as 'using' table when creating a delete query
        clauses = list(chain(*[self._attribute(a, mapping.name) for a in mapping.attributes if a.unique or not self._is_delete_query]))

        return ' and '.join(clauses)

    def _attribute(self, attribute: AbstractAttribute, alias) -> List[str]:
        if isinstance(attribute, Attribute):
            # no where-clause needed for root table. Only add where clauses for joined tables.
            return []

        if isinstance(attribute, Reference):

            # don't add extension conditions if this is an insert query
            if attribute.extension and self._is_insert_query:
                return []

            # recurse
            target_alias = attribute.alias or attribute.table
            where_clause = self._nested_attributes(attribute.attributes, target_alias)

            # add extension conditions if this is the root of an update or delete query
            if attribute.extension and attribute.qualifier and not self._is_insert_query:
                # assume for now that the alias is the table name. This is fine as long as we're not joining the same table multiple times
                model = alias.replace('_', '.')
                extension_clause = [f"{target_alias}.module = '{attribute.qualifier}'",
                                    f"{target_alias}.model = '{model}'"]
                where_clause = chain(where_clause, extension_clause)

            return where_clause

    def _nested_attribute(self, attribute: AbstractAttribute, alias):
        if isinstance(attribute, Attribute):

            # terminate foreign key with an equation
            parameter_name = attribute.parameter or attribute.name

            # if there's a key, then address field in json object
            if attribute.key:
                return [f"{alias}.{attribute.name}->>'{attribute.key}' = :{parameter_name}"]

            return [f'{alias}.{attribute.name} = :{parameter_name}']

        if isinstance(attribute, Reference):
            # recurse
            target_alias = attribute.alias or attribute.table
            where_clause = self._nested_attributes(attribute.attributes, target_alias)

            return where_clause

    def _nested_attributes(self, attributes: List[AbstractAttribute], alias):
        return list(chain(*[self._nested_attribute(attribute, alias) for attribute in attributes]))



"""
This script defines an Invoker class intended for local invocations of the Stimula API.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
import sys

import jwt
import psycopg2
from sqlalchemy import create_engine

from stimula.service.abstract_orm import AbstractORM
from stimula.service.auth import Auth
from stimula.service.context import cnx_context
from stimula.service.db import DB


class Invoker:
    def __init__(self, secret_key, host, port):
        # for local client, the secret key does not depend on the database but is specified by the user
        self._auth = LocalAuth(lambda database: secret_key, host, port)
        self._db = DB(lambda: MuteORM())

    def set_context(self, token):
        # set context for processing of this request
        cnx_context.cnx, cnx_context.cr, cnx_context.username = self._auth.validate_token(token)

        # also create sqlalchemy engine, bec/ that's what pandas needs
        cnx_context.engine = create_engine('postgresql://', creator=lambda: cnx_context.cnx)

    def auth(self, database, username, password):
        return self._auth.authenticate(database, username, password)

    def get_database_and_username(self, token):
        return self._auth.get_database_and_username(token)

    def list(self, filter):
        return self._db.get_tables(filter)

    def mapping(self, table):
        return self._db.get_header_csv(table)

    def count(self, table, header, query):
        return self._db.get_count(table, header, query)

    def get_table(self, table, header, query):
        return self._db.get_table_as_csv(table, header, query)

    def post_table(self, table, header, query, files, skiprows, nrows, block_size, insert, update, delete, execute, commit, format, post_script, context, substitutions):

        if format == None or format == 'diff':
            # post table and get diff dataframes
            post_result = self._db.post_table_get_diff(table, header, query, files, skiprows=skiprows, nrows=nrows, insert=insert, update=update, delete=delete, execute=execute, commit=commit, post_script=post_script)
            # convert dataframes to CSV response body
            return '\n\n'.join([df.to_csv(index=False) for df in post_result])
        elif format == 'sql':
            # post table and get sql
            post_result = self._db.post_table_get_sql(table, header, query, files, skiprows=skiprows, nrows=nrows, insert=insert, update=update, delete=delete, execute=execute, commit=commit, post_script=post_script)
            # convert df to response body, use double quotes where needed
            return post_result.to_csv(index=False, quotechar="\"")
        elif format == 'full' and len(files) == 1:
            assert len(table) == 1, "Provide exactly one table name to match contents, not %s" % len(table)
            # take body from first file and convert to string
            body = files[0].decode('utf-8')
            context = context[0] if context and len(context) == 1 else None
            substitutions = substitutions[0].decode('utf-8') if substitutions[0] else None
            # post table and get full report
            post_result = self._db.post_table_get_full_report(table[0], header, query, body, skiprows=skiprows, nrows=nrows, insert=insert, update=update, delete=delete, execute=execute, commit=commit, post_script=post_script, context=context, substitutions=substitutions)
            # return json as string
            return post_result
        elif len(files) > 1:
            post_result = self._db.post_multiple_tables_get_full_report(table, header, query, files, skiprows=skiprows, nrows=nrows, insert=insert, update=update, delete=delete, execute=execute, commit=commit, post_script=post_script, context=context)
            return post_result


class LocalAuth(Auth):
    # set the secret key during instantiation
    def __init__(self, secret_key_function, host, port):
        super().__init__(secret_key_function)
        self._host = host
        self._port = port

    def _validate_submitted_credentials(self, database, username, password):
        # create connection url
        url = f"postgresql://{username}@{self._host}:{self._port}/{database}"

        # use psycopg2 to connect to the database
        cnx = psycopg2.connect(url, password=password)

        # create cursor
        cr = cnx.cursor()

        # test connection
        cr.execute("SELECT 1")

        return username

    def _validate_token_credentials(self, database, uid, password):
        # use psycopg2 to connect to the database
        cnx = psycopg2.connect(dbname=database, user=uid, password=password, host=self._host, port=self._port)

        # create cursor
        cr = cnx.cursor()

        # test connection
        cr.execute("SELECT 1")

        # return connection and cursor objects. With psycopg2, it's best to re-use the connection and cursor objects
        return cnx, cr

class MuteORM(AbstractORM):
    def create(self, model_name: str, values: dict):
        pass

    def read(self, model_name: str, record_id: int):
        pass

    def update(self, model_name: str, record_id: int, values: dict):
        pass

    def delete(self, model_name: str, record_id: int):
        pass



# -----------------------------------------------------------------------------
# sly: lex.py
#
# Copyright (C) 2016 - 2018
# David M. Beazley (Dabeaz LLC)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of the David Beazley or Dabeaz LLC may be used to
#   endorse or promote products derived from this software without
#  specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

__all__ = ['Lexer', 'LexerStateChange']

import re
import copy

class LexError(Exception):
    '''
    Exception raised if an invalid character is encountered and no default
    error handler function is defined.  The .text attribute of the exception
    contains all remaining untokenized text. The .error_index is the index
    location of the error.
    '''
    def __init__(self, message, text, error_index):
        self.args = (message,)
        self.text = text
        self.error_index = error_index

class PatternError(Exception):
    '''
    Exception raised if there's some kind of problem with the specified
    regex patterns in the lexer.
    '''
    pass

class LexerBuildError(Exception):
    '''
    Exception raised if there's some sort of problem building the lexer.
    '''
    pass

class LexerStateChange(Exception):
    '''
    Exception raised to force a lexing state change
    '''
    def __init__(self, newstate, tok=None):
        self.newstate = newstate
        self.tok = tok

class Token(object):
    '''
    Representation of a single token.
    '''
    __slots__ = ('type', 'value', 'lineno', 'index', 'end')
    def __repr__(self):
        return f'Token(type={self.type!r}, value={self.value!r}, lineno={self.lineno}, index={self.index}, end={self.end})'

class TokenStr(str):
    @staticmethod
    def __new__(cls, value, key=None, remap=None):
        self = super().__new__(cls, value)
        self.key = key
        self.remap = remap
        return self

    # Implementation of TOKEN[value] = NEWTOKEN
    def __setitem__(self, key, value):
        if self.remap is not None:
            self.remap[self.key, key] = value

    # Implementation of del TOKEN[value]
    def __delitem__(self, key):
        if self.remap is not None:
            self.remap[self.key, key] = self.key

class _Before:
    def __init__(self, tok, pattern):
        self.tok = tok
        self.pattern = pattern

class LexerMetaDict(dict):
    '''
    Special dictionary that prohibits duplicate definitions in lexer specifications.
    '''
    def __init__(self):
        self.before = { }
        self.delete = [ ]
        self.remap = { }

    def __setitem__(self, key, value):
        if isinstance(value, str):
            value = TokenStr(value, key, self.remap)
            
        if isinstance(value, _Before):
            self.before[key] = value.tok
            value = TokenStr(value.pattern, key, self.remap)
            
        if key in self and not isinstance(value, property):
            prior = self[key]
            if isinstance(prior, str):
                if callable(value):
                    value.pattern = prior
                else:
                    raise AttributeError(f'Name {key} redefined')

        super().__setitem__(key, value)

    def __delitem__(self, key):
        self.delete.append(key)
        if key not in self and key.isupper():
            pass
        else:
            return super().__delitem__(key)

    def __getitem__(self, key):
        if key not in self and key.split('ignore_')[-1].isupper() and key[:1] != '_':
            return TokenStr(key, key, self.remap)
        else:
            return super().__getitem__(key)

class LexerMeta(type):
    '''
    Metaclass for collecting lexing rules
    '''
    @classmethod
    def __prepare__(meta, name, bases):
        d = LexerMetaDict()

        def _(pattern, *extra):
            patterns = [pattern, *extra]
            def decorate(func):
                pattern = '|'.join(f'({pat})' for pat in patterns )
                if hasattr(func, 'pattern'):
                    func.pattern = pattern + '|' + func.pattern
                else:
                    func.pattern = pattern
                return func
            return decorate

        d['_'] = _
        d['before'] = _Before
        return d

    def __new__(meta, clsname, bases, attributes):
        del attributes['_']
        del attributes['before']

        # Create attributes for use in the actual class body
        cls_attributes = { str(key): str(val) if isinstance(val, TokenStr) else val
                           for key, val in attributes.items() }
        cls = super().__new__(meta, clsname, bases, cls_attributes)

        # Attach various metadata to the class
        cls._attributes = dict(attributes)
        cls._remap = attributes.remap
        cls._before = attributes.before
        cls._delete = attributes.delete
        cls._build()
        return cls

class Lexer(metaclass=LexerMeta):
    # These attributes may be defined in subclasses
    tokens = set()
    literals = set()
    ignore = ''
    reflags = 0
    regex_module = re

    _token_names = set()
    _token_funcs = {}
    _ignored_tokens = set()
    _remapping = {}
    _delete = {}
    _remap = {}

    # Internal attributes
    __state_stack = None
    __set_state = None

    @classmethod
    def _collect_rules(cls):
        # Collect all of the rules from class definitions that look like token
        # information.   There are a few things that govern this:
        #
        # 1.  Any definition of the form NAME = str is a token if NAME is
        #     is defined in the tokens set.
        #
        # 2.  Any definition of the form ignore_NAME = str is a rule for an ignored
        #     token.
        #
        # 3.  Any function defined with a 'pattern' attribute is treated as a rule.
        #     Such functions can be created with the @_ decorator or by defining
        #     function with the same name as a previously defined string.
        #
        # This function is responsible for keeping rules in order. 

        # Collect all previous rules from base classes
        rules = []

        for base in cls.__bases__:
            if isinstance(base, LexerMeta):
                rules.extend(base._rules)
                
        # Dictionary of previous rules
        existing = dict(rules)

        for key, value in cls._attributes.items():
            # strange fix, on docker, regex_module comes with 'pattern' attribute, even though re.py is the same as on local machine
            # exclude it, because it's not a patter and causes a error later on
            # if (key in cls._token_names) or key.startswith('ignore_') or hasattr(value, 'pattern'):
            if (key in cls._token_names) or key.startswith('ignore_') or (hasattr(value, 'pattern') and key != 'regex_module'):
                if callable(value) and not hasattr(value, 'pattern'):
                    raise LexerBuildError(f"function {value} doesn't have a regex pattern")
                
                if key in existing:
                    # The definition matches something that already existed in the base class.
                    # We replace it, but keep the original ordering
                    n = rules.index((key, existing[key]))
                    rules[n] = (key, value)
                    existing[key] = value

                elif isinstance(value, TokenStr) and key in cls._before:
                    before = cls._before[key]
                    if before in existing:
                        # Position the token before another specified token
                        n = rules.index((before, existing[before]))
                        rules.insert(n, (key, value))
                    else:
                        # Put at the end of the rule list
                        rules.append((key, value))
                    existing[key] = value

                # fix: ignore non callable values such as a module, because it does not contain Lexer tokens
                # elif callable(value):
                else:
                    rules.append((key, value))
                    existing[key] = value

            elif isinstance(value, str) and not key.startswith('_') and key not in {'ignore', 'literals'}:
                raise LexerBuildError(f'{key} does not match a name in tokens')

        # Apply deletion rules
        rules = [ (key, value) for key, value in rules if key not in cls._delete ]
        cls._rules = rules

    @classmethod
    def _build(cls):
        '''
        Build the lexer object from the collected tokens and regular expressions.
        Validate the rules to make sure they look sane.
        '''
        if 'tokens' not in vars(cls):
            raise LexerBuildError(f'{cls.__qualname__} class does not define a tokens attribute')

        # Pull definitions created for any parent classes
        cls._token_names = cls._token_names | set(cls.tokens)
        cls._ignored_tokens = set(cls._ignored_tokens)
        cls._token_funcs = dict(cls._token_funcs)
        cls._remapping = dict(cls._remapping)

        for (key, val), newtok in cls._remap.items():
            if key not in cls._remapping:
                cls._remapping[key] = {}
            cls._remapping[key][val] = newtok

        remapped_toks = set()
        for d in cls._remapping.values():
            remapped_toks.update(d.values())
            
        undefined = remapped_toks - set(cls._token_names)
        if undefined:
            missing = ', '.join(undefined)
            raise LexerBuildError(f'{missing} not included in token(s)')

        cls._collect_rules()

        parts = []
        for tokname, value in cls._rules:
            if tokname.startswith('ignore_'):
                tokname = tokname[7:]
                cls._ignored_tokens.add(tokname)

            if isinstance(value, str):
                pattern = value

            elif callable(value):
                cls._token_funcs[tokname] = value
                pattern = getattr(value, 'pattern')

            # Form the regular expression component
            part = f'(?P<{tokname}>{pattern})'

            # Make sure the individual regex compiles properly
            try:
                cpat = cls.regex_module.compile(part, cls.reflags)
            except Exception as e:
                raise PatternError(f'Invalid regex for token {tokname}') from e

            # Verify that the pattern doesn't match the empty string
            if cpat.match(''):
                raise PatternError(f'Regex for token {tokname} matches empty input')

            parts.append(part)

        if not parts:
            return

        # Form the master regular expression
        #previous = ('|' + cls._master_re.pattern) if cls._master_re else ''
        # cls._master_re = cls.regex_module.compile('|'.join(parts) + previous, cls.reflags)
        cls._master_re = cls.regex_module.compile('|'.join(parts), cls.reflags)

        # Verify that that ignore and literals specifiers match the input type
        if not isinstance(cls.ignore, str):
            raise LexerBuildError('ignore specifier must be a string')

        if not all(isinstance(lit, str) for lit in cls.literals):
            raise LexerBuildError('literals must be specified as strings')

    def begin(self, cls):
        '''
        Begin a new lexer state
        '''
        assert isinstance(cls, LexerMeta), "state must be a subclass of Lexer"
        if self.__set_state:
            self.__set_state(cls)
        self.__class__ = cls

    def push_state(self, cls):
        '''
        Push a new lexer state onto the stack
        '''
        if self.__state_stack is None:
            self.__state_stack = []
        self.__state_stack.append(type(self))
        self.begin(cls)

    def pop_state(self):
        '''
        Pop a lexer state from the stack
        '''
        self.begin(self.__state_stack.pop())

    def tokenize(self, text, lineno=1, index=0):
        _ignored_tokens = _master_re = _ignore = _token_funcs = _literals = _remapping = None

        # --- Support for state changes
        def _set_state(cls):
            nonlocal _ignored_tokens, _master_re, _ignore, _token_funcs, _literals, _remapping
            _ignored_tokens = cls._ignored_tokens
            _master_re = cls._master_re
            _ignore = cls.ignore
            _token_funcs = cls._token_funcs
            _literals = cls.literals
            _remapping = cls._remapping

        self.__set_state = _set_state
        _set_state(type(self))

        # --- Support for backtracking
        _mark_stack = []
        def _mark():
            _mark_stack.append((type(self), index, lineno))
        self.mark = _mark

        def _accept():
            _mark_stack.pop()
        self.accept = _accept

        def _reject():
            nonlocal index, lineno
            cls, index, lineno = _mark_stack[-1]
            _set_state(cls)
        self.reject = _reject


        # --- Main tokenization function
        self.text = text
        try:
            while True:
                try:
                    if text[index] in _ignore:
                        index += 1
                        continue
                except IndexError:
                    return

                tok = Token()
                tok.lineno = lineno
                tok.index = index
                m = _master_re.match(text, index)
                if m:
                    tok.end = index = m.end()
                    tok.value = m.group()
                    tok.type = m.lastgroup

                    if tok.type in _remapping:
                        tok.type = _remapping[tok.type].get(tok.value, tok.type)

                    if tok.type in _token_funcs:
                        self.index = index
                        self.lineno = lineno
                        tok = _token_funcs[tok.type](self, tok)
                        index = self.index
                        lineno = self.lineno
                        if not tok:
                            continue

                    if tok.type in _ignored_tokens:
                        continue

                    yield tok

                else:
                    # No match, see if the character is in literals
                    if text[index] in _literals:
                        tok.value = text[index]
                        tok.end = index + 1
                        tok.type = tok.value
                        index += 1
                        yield tok
                    else:
                        # A lexing error
                        self.index = index
                        self.lineno = lineno
                        tok.type = 'ERROR'
                        tok.value = text[index:]
                        tok = self.error(tok)
                        if tok is not None:
                            tok.end = self.index
                            yield tok

                        index = self.index
                        lineno = self.lineno

        # Set the final state of the lexer before exiting (even if exception)
        finally:
            self.text = text
            self.index = index
            self.lineno = lineno

    # Default implementations of the error handler. May be changed in subclasses
    def error(self, t):
        raise LexError(f'Illegal character {t.value[0]!r} at index {self.index}', t.value, self.index)



import base64

import requests

from stimula.stml.model import Attribute

'''
This class reads documents from remote APIs.
'''


class ApiReader:
    def read_document(self, attribute: Attribute, params):
        # get url, expand placeholders with values from row
        url = attribute.url.format(**params)

        # create authorization header
        headers = {"Authorization": attribute.auth} if attribute.auth else {}

        # send http GET request to url
        response = requests.get(url, headers=headers)

        # assert that the response status code is 200
        assert response.status_code == 200, f"API call to {url} failed with status code {response.status_code}"

        # check if the response is an AFAS document
        if self._is_afas_document(response):
            # extract afas document from response
            return self._extract_afas_document(response, params)

        # return response content
        return response.content

    def _is_afas_document(self, response):
        # check if the response is an AFAS document
        return 'application/json' in response.headers.get('Content-Type', '') and 'filedata' in response.json()

    def _extract_afas_document(self, response, params):
        #     extract json from response
        response_json = response.json()

        # verify document name
        assert 'filename' in response_json, "AFAS response does not contain a filename"

        # get filename from response
        filename = response_json['filename']

        # get expected name from params
        expected_name = params['name']

        assert filename == expected_name, f"AFAS document name {filename} does not match expected name {expected_name}"
        assert 'filedata' in response_json, "AFAS response does not contain filedata"

        # decode filedata from base64
        document = base64.b64decode(response_json['filedata'])

        # return document
        return document



import random
from io import TextIOBase

import pandas as pd


class Anonymizer:
    def anonymize(self, input, output):
        # read dataframe from stdin, read all as str to avoid conversion errors
        df = pd.read_csv(input, dtype=str)

        # get columns to anonymize
        columns = self.get_sensitive_columns(df)

        # anonymize columns
        df = self.anonymize_columns(df, columns)

        # write to stdout
        df.to_csv(output, index=False)

        return df

    def get_sensitive_columns(self, df):
        # filter sensitive column names
        return [col for col in df.columns if self.is_sensitive(col, df[col])]

    def is_sensitive(self, name, values):
        # list non-empty values
        non_empty_value_count = len(values.dropna())

        # if none left, return false
        if non_empty_value_count == 0:
            return False

        # if all non-empty values are numeric, then return false
        if pd.to_numeric(values.dropna(), errors='coerce').notnull().all():
            return False

        # lower case the column name
        name = name.lower()

        # if the column contains the word "name", "address" or "email", then return true
        if 'name' in name or 'address' in name or 'email' in name:
            return True

        if 'zip' in name or 'city' in name or 'date' in name:
            return False

        # if too few values provided, then return true
        if non_empty_value_count < 4:
            return True

        # count unique values
        unique_count = len(values.unique())

        # if unique values are less than or equal to 25% of the total values, then return false
        if unique_count <= 0.25 * non_empty_value_count:
            return False

        # else don't know, return true
        return True

    def anonymize_columns(self, df, columns):
        for col in columns:
            df[col] = df[col].apply(lambda x: self.anonymize_value(x))
        return df

    def anonymize_value(self, x):
        # if nan, return as is
        if pd.isnull(x):
            return x

        # if x is not a string, return as is
        if not isinstance(x, str):
            return x

        # split words by space
        words = x.split()

        # anonymize each word
        return ' '.join(self.anonymize_word(word) for word in words)

    def anonymize_word(self, x):
        # is the word capitalized?
        is_capitalized = len(x) > 1 and x[0].isupper() and x[1:].islower()
        # n is the number of times to swap characters. Minimum must be an odd number to minimize the chance of the word left unchanged
        n = max(3, len(x) // 8)
        # iterate n times
        for i in range(n):
            # create two random numbers, smaller than the length of the string
            i = random.randint(0, len(x) - 1)
            j = random.randint(0, len(x) - 1)
            # swap two random characters
            x = self.swap(x, i, j)

        # if the word was capitalized, capitalize it again
        if is_capitalized:
            x = x.capitalize()
        return x

    def swap(self, s, i, j):
        l = list(s)
        l[i], l[j] = l[j], l[i]
        return ''.join(l)



import base64
import hashlib
import importlib
import logging
import os
import re
import sys
from io import StringIO

import pandas as pd

from stimula.service.api_reader import ApiReader
from stimula.stml.header_renderer import HeaderRenderer
from stimula.stml.model import Entity, Attribute
from stimula.stml.sql.types_renderer import TypesRenderer

_logger = logging.getLogger(__name__)


class CsvReader:
    def read_from_request(self, mapping, body, skiprows, nrows=None, post_script=None, substitutions_map: dict = None):

        # get columns and unique columns. Include all columns, including skip and orm-only columns.
        column_names = HeaderRenderer().render_list(mapping, include_skip=True, include_orm_only=True)
        index_columns = HeaderRenderer().render_list_unique(mapping)
        column_types = TypesRenderer().render(mapping, column_names, include_skip=True, include_orm_only=True, substitutions=substitutions_map)
        deduplicate_columns = HeaderRenderer().render_list_deduplicate(mapping)

        # assert that at least one column header is not empty
        if not [c for c in column_names if c != '']:
            raise ValueError("At least one column header must not be empty")

        # read body as csv to count the number of columns in the body
        # there may be a more efficient way without having to read the csv twice.
        body_column_count = self._count_body_columns(body)

        # pad column names with empty columns if the body has more columns than the header
        if body_column_count > len(column_names):
            # add empty columns to the header
            column_names += [''] * (body_column_count - len(column_names))

        # replace empty column names with skip, skip1, skip2. This is because pandas requires column names to be unique
        non_empty_column_names = list(self._replace_empty_columns_with_skip(column_names))

        # find duplicate column names
        duplicate_column_names = self._find_duplicate_names(non_empty_column_names)

        # if there are duplicate column names, raise an exception
        if duplicate_column_names:
            raise ValueError(f"Duplicate column names are not supported: {', '.join(duplicate_column_names)}")

        # get list of columns to use in the output dataframe
        use_columns = [c for c in non_empty_column_names if c in column_names]

        # list names of columns with datetime64 or date type, because we need to parse them as datetime
        parse_dates = column_types.get('read_csv_parse_dates', {})

        # get converter dictionary
        converters = column_types.get('read_csv_converters', {})

        # get dtypes for read_csv
        dtype = column_types.get('read_csv_dtypes', {})

        # create initial column names to read the csv before padding
        initial_index_columns = [c for c in index_columns if c in non_empty_column_names[:body_column_count]]
        initial_names = non_empty_column_names[:body_column_count]
        initial_usecols = [c for c in use_columns if c in non_empty_column_names[:body_column_count]]

        # read csv from request body
        # treat '' as missing value, but treat NA as string
        df = pd.read_csv(
            StringIO(body),
            names=initial_names,
            index_col=initial_index_columns,
            skipinitialspace=True,
            skiprows=skiprows,
            nrows=nrows,
            # skip the empty column names that we renamed to skip, skip1, etc
            usecols=initial_usecols,
            parse_dates=parse_dates,
            converters=converters,
            na_values=[''],
            keep_default_na=False
        )

        # Restore index and column types, because pd set converter results to type object.
        self._restore_column_types(df, dtype)

        # pad dataframe with empty columns if we have more column names in use_columns than exist in the dataframe
        df_padded = self._pad_dataframe_with_empty_columns(df, use_columns, index_columns, converters)

        # filter rows based on expression
        self._filter_rows(df_padded, mapping, column_names, index_columns)

        # invoke apis to retrieve additional data, such as attachments
        self._invoke_apis(df_padded, mapping, index_columns)

        # evaluate column expressions, must do after restoring index and column types
        self._evaluate_expressions(df_padded, mapping, column_names, index_columns)

        # insert a column with line numbers
        df_padded.insert(0, '__line__', range(0, len(df)))

        # deduplicate if requested
        if deduplicate_columns:
            df_padded = self._deduplicate(df_padded, deduplicate_columns)

        # verify that there are no duplicate index values
        if df_padded.index.has_duplicates:
            # find duplicate index values
            duplicates = df_padded.index[df_padded.index.duplicated()]
            # convert to name-value dict
            duplicate_map = {k: v for k, v in zip(duplicates.names, duplicates.values)}

            raise ValueError(f"Duplicates found: {duplicate_map}")

        # apply post script if provided
        if post_script:
            # execute post script
            self._execute_post_script(df_padded, post_script)

        return df_padded

    def _count_body_columns(self, body):
        # skipinitialspace must be true, otherwise it may split on comma's in strings
        df_initial = pd.read_csv(StringIO(body), skipinitialspace=True, header=None)

        # return the number of columns in the body
        return len(df_initial.columns)

    def _replace_empty_columns_with_skip(self, column_names):
        # replace empty column names with skip, skip1, skip2. This is because pandas doesn't like empty column names
        i = 0
        for name in column_names:
            if name == '':
                name = 'skip' + ('' if i == 0 else str(i))
                i += 1
            yield name

    def _find_duplicate_names(self, names):
        # find duplicate names
        seen = set()
        duplicates = set()
        for name in names:
            if name in seen:
                duplicates.add(name)
            seen.add(name)
        return duplicates

    def _restore_column_types(self, df, dtype):
        # iterate columns types and convert to the correct type.
        # Need to do this after reading, because pd sets converter results to type object
        # Need to do this before evaluating expressions, because numexpr (if installed) can't deal with type 'object'
        for column, type in dtype.items():
            # skip setting index types
            if column in df.columns:
                # convert column to the correct type
                df[column] = df[column].astype(type)
            elif column in df.index.names and isinstance(df.index, pd.MultiIndex):
                # a jsonb column has type object, but pandas does not support setting a multi-index to type object
                if not type == 'object':
                    # convert multi-index to the correct type
                    df.index = df.index.set_levels(df.index.levels[df.index.names.index(column)].astype(type), level=column)
            elif column in df.index.names and not isinstance(df.index, pd.MultiIndex):
                # convert single-level index to the correct type
                df.index = df.index.astype(type)

    def _pad_dataframe_with_empty_columns(self, df, column_names, index_columns, converters):
        # get current column names
        current_columns = df.columns.tolist()

        # get number of columns to pad, take columns and indices into account
        pad_count = len(column_names) - df.shape[1] - df.index.nlevels

        # if there are too few columns
        if pad_count <= 0:
            # nothing to pad
            return df

        # get columns to pad
        pad_columns = column_names[-pad_count:]

        # pad the DataFrame with empty columns
        _logger.info(f"Padding DataFrame with {pad_count} empty columns")
        current_columns += pad_columns

        # be careful not to set index columns, the columns attribute must only be set to non-index column names
        df_padded = df.reindex(columns=current_columns, fill_value=None)

        # apply converters to padded columns
        for column in pad_columns:
            if column in converters:
                df_padded[column] = df_padded[column].apply(converters[column])

        # set additional index columns
        for column in pad_columns:
            if column in index_columns:
                df_padded.set_index(column, append=True, inplace=True)

        return df_padded

    def _filter_rows(self, df, mapping, column_names, index_columns):
        # A column header can contain a python expression to filter rows.
        # Evaluate these now that we've read all values from CSV, but before invoking APIs and evaluating expressions.

        # drop index, so that we can use index columns in expressions
        if index_columns:
            df.reset_index(inplace=True)

        # store original column names so we can restore them later
        original_column_names = df.columns

        # remove foreign keys and modifiers in column names so we can use the bare names in expressions
        df.columns = df.columns.str.replace(r'\[.*\]', '', regex=True)
        df.columns = df.columns.str.replace(r'\(.*\)', '', regex=True)

        # now that we have all data, we can evaluate column expressions.
        for attribute in mapping.attributes:
            if attribute and attribute.filter_src:
                # query to filter rows
                df.query(attribute.filter_src, inplace=True)

        # restore column names
        df.columns = original_column_names

        # restore index
        if index_columns:
            df.set_index(index_columns, inplace=True)

    def _invoke_apis(self, df, mapping: Entity, index_columns):
        # a column header can contain an 'api' modifier. Invoke these now that we've read all values from CSV, but before evaluating expressions.

        # drop index, so that we can use index columns in expressions
        if index_columns:
            df.reset_index(inplace=True)

        # store original column names so we can restore them later
        original_column_names = df.columns

        # remove foreign keys and modifiers in column names so we can use the bare names in expressions
        df.columns = df.columns.str.replace(r'\[.*\]', '', regex=True)
        df.columns = df.columns.str.replace(r'\(.*\)', '', regex=True)

        # now that we have all data, we can evaluate api calls.
        for attribute in mapping.attributes:
            if isinstance(attribute, Attribute) and attribute.api:
                # get column name, assuming a single attribute
                column_name = attribute.name

                # assert that the column has a url modifier
                assert attribute.url, f"Column {column_name} has an 'api' modifier, but no 'url' modifier"

                # drop all values in the column and set type to binary
                df[column_name] = None
                df[column_name] = df[column_name].astype('bytes')

                # iterate rows and invoke the api
                for index, row in df.iterrows():
                    try:
                        # read document and add the binary response to the row in the DataFrame
                        df.at[index, column_name] = ApiReader().read_document(attribute, row.to_dict())
                    except Exception as e:
                        # log error and continue to next row
                        _logger.error(f"Error invoking API for row {index}: {e}")

        # restore column names
        df.columns = original_column_names

        # restore index
        if index_columns:
            df.set_index(index_columns, inplace=True)

    def _evaluate_expressions(self, df, mapping, column_names, index_columns):
        # a column header can contain a python expression. Evaluate these now that we've read all values from CSV.

        # drop index, so that we can use index columns in expressions
        if index_columns:
            df.reset_index(inplace=True)

        # store original column names so we can restore them later
        original_column_names = df.columns

        # remove foreign keys and modifiers in column names so we can use the bare names in expressions
        df.columns = df.columns.str.replace(r'\[.*\]', '', regex=True)
        df.columns = df.columns.str.replace(r'\(.*\)', '', regex=True)

        # now that we have all data, we can evaluate column expressions.
        for attribute in mapping.attributes:
            if attribute and attribute.exp:
                # get column name, assuming a single attribute
                column_name = attribute.name

                # evaluate expression of the form <target_column>=<expression>
                expression = f"{column_name}={attribute.exp}"

                # evaluate the expression, pass custom functions
                df.eval(expression, inplace=True, local_dict={'checksum': checksum, 'base64encode': base64encode, 'concat': concat, 'fallback': fallback, 'afas_file_name': afas_file_name})

        # restore column names
        df.columns = original_column_names

        # restore index
        if index_columns:
            df.set_index(index_columns, inplace=True)

        # list all columns with 'skip=true' in their mapping, but not API results. Assume single attribute.
        drop_column_names = [n for n, a in zip(column_names, mapping.attributes) if a and a.skip and not (isinstance(a, Attribute) and a.api)]

        # drop these columns, because we've evaluated expressions so we no longer need them. But keep API results, we'll use them later.
        df.drop(columns=drop_column_names, errors='ignore', inplace=True)

    def _deduplicate(self, df, index_columns_to_deduplicate):
        # get original index columns
        original_index_columns = df.index.names

        # reset index
        df_reset = df.reset_index()

        # Remove duplicate rows based on the index column
        df_unique = df_reset.drop_duplicates(subset=index_columns_to_deduplicate)

        # Set the column back as the index
        df_final = df_unique.set_index(original_index_columns)

        return df_final

    def _execute_post_script(self, df, post_script):
        # skip if no post script provided
        if post_script is None:
            return df

        # assert that post_script file exists
        assert os.path.exists(post_script), f"Post script file {post_script} not found"

        # import the post script module
        module_name = 'post_script'
        spec = importlib.util.spec_from_file_location(module_name, post_script)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)

        # verify that the module has an execute function
        assert hasattr(module, 'execute'), f"Post script module {post_script} must have an execute function"

        # execute the post script
        return module.execute(df)


def checksum(series):
    # checksum function for custom expression. Return hex digest for all items in series an return as type string.
    return series.apply(_hexdigest).astype('string')


def _hexdigest(x):
    # if x is None, return None
    if pd.isna(x):
        return None
    # if x is str, encode and return hex digest
    if isinstance(x, str):
        return hashlib.sha1(x.encode()).hexdigest()
    # if x is bytes, return hex digest
    if isinstance(x, bytes):
        return hashlib.sha1(x).hexdigest()
    # else raise an exception
    raise ValueError(f"Unsupported type {type(x)}")


def base64encode(series):
    # UU-encode function for custom expression.
    return series.apply(_base64encode).astype('string')


def _base64encode(x: str | bytes | None) -> str | None:
    # Return None if the input is None
    if x is None:
        return None

    # If the input is a string, encode it to bytes and Base64-encode
    if isinstance(x, str):
        return base64.b64encode(x.encode()).decode('utf-8')

    # If the input is bytes, Base64-encode it directly
    if isinstance(x, bytes):
        return base64.b64encode(x).decode('utf-8')

    # Raise an exception for unsupported types
    raise ValueError(f"Unsupported type {type(x)}")


def concat(*series):
    # copy tuple into a list
    series_list = list(series)

    # concat function for custom expression. Return concatenated string for all items in series. Return as type 'string'.
    sep = series_list[0] or ''
    enc = '"' if bool(series_list[1]) else ''

    # get the length of the longest series
    max_len = max(len(s) for s in series_list if isinstance(s, pd.Series))

    # expand fixed elements to series of the same value, so we can also concat constant strings
    for i in range(2, len(series_list)):
        if not isinstance(series_list[i], pd.Series):
            # replace with a series of the same value, with length max_len
            series_list[i] = pd.Series([series_list[i]] * max_len)

    # apply concatenation to all rows
    return pd.Series(list(zip(*series_list[2:]))).apply(lambda s: _concat(sep, enc, s)).astype('string')


# a concat function that accepts 1+ arguments and returns a concatenated string, using first parameter as separator. If second parameter is true, then enclose in quotes
def _concat(sep, enc, series):
    # concatenate non-empty items in series, using sep as separator and dec as enclosure
    return sep.join(enc + str(s) + enc for s in series if not pd.isna(s) and not s == '') or ''


def fallback(*series):
    # takes any number of values and returns the first non-null and non-empty value
    return pd.Series(list(zip(*series))).apply(_fallback).astype('string')


def _fallback(args):
    for arg in args:
        if not pd.isna(arg) and not arg == '':
            return arg
    return ''


def afas_file_name(*series):
    # takes any number of values and returns the first non-null and non-empty value
    return pd.Series(list(zip(*series))).apply(_afas_file_name).astype('string')


def _afas_file_name(args):
    assert len(args) == 1, f"Expected 1 argument, got {args}"
    afas_encoding_map = {
        '/': '_2F',
        '#': '_23',
        '&': '_26',
        ':': '_3A',
        '?': '_3F',
        '*': '_2A',
        '<': '_3C',
        '>': '_3E',
        '%': '_25',
        '+': '_2B',
        '~': '_7E',
        '-': '_2D',
        '@': '_40',
        '!': '_21',
        '$': '_24',
        '_': '_5F',
        '\'': '_27',
    }
    if args[0]:
        # replace special characters with their afas encoding
        afas_name = re.sub(r"[/#&:?*<>%+~\-@!$_']", lambda m: afas_encoding_map[m.group()], args[0])
        return afas_name

    return ''



import pandas as pd

from stimula.service.model_service import ModelService
from stimula.service.odoo.jsonrpc_model_service import JsonRpcModelService
from stimula.service.odoo.postgres_model_service import PostgresModelService
from stimula.stml.header_renderer import HeaderRenderer
from stimula.stml.sql.types_renderer import TypesRenderer

MODEL_SERVICES = {
    "sql": PostgresModelService,
    "jsonrpc": JsonRpcModelService
}


class DbReader:
    def __init__(self, protocol='sql'):
        assert protocol in MODEL_SERVICES, f"Protocol '{protocol}' not supported"
        self._model_service: ModelService = MODEL_SERVICES[protocol]()

    def read_from_db(self, mapping, where_clause, set_index=False):

        # get enabled and unique columns and column types
        column_names = HeaderRenderer().render_list(mapping)
        index_columns = HeaderRenderer().render_list_unique(mapping)
        column_types = TypesRenderer().render(mapping, column_names)

        # read dataframe from DB
        df = self._model_service.read_table(mapping, where_clause)

        # set headers, they must equal the request headers for comparison
        df.columns = column_names

        # apply converters after reading from DB, because read_sql_query() doesn't support converters
        converters = column_types['read_db_converters']
        self._apply_converters(column_names, converters, df)

        # force pandas to not convert int columns to float if they contain NaNs
        df = df.convert_dtypes()

        # set index columns, if header contains unique columns. Set index after converting, because we need to convert dict to frozenset so it is hashable.
        if set_index and index_columns:
            df.set_index(index_columns, inplace=True)

        return df

    def _apply_converters(self, column_names, converters, df):
        # iterate the columns to apply converters one by one to either value or index column
        for column in column_names:
            if column in converters:
                if column in df.columns:

                    # apply this converter to value column
                    df[column] = df[column].apply(converters[column])
                elif column in df.index.names:
                    if isinstance(df.index, pd.MultiIndex):

                        # get the index column for a MultiIndex
                        index_column = df.index.levels[df.index.names.index(column)]

                        # apply this converter to the index column for a MultiIndex
                        df.index = df.index.set_levels(index_column.map(converters[column]), level=column)
                    else:
                        # if it's not a MultiIndex, just apply the converter to the single-level index
                        df.index = df.index.map(converters[column])



import io
import os
import re
import sys

import pandas as pd
from numpy import isnan


class FileSource:
    def read_files(self, file_names, table_names, context):
        '''
        Read file contents from disk or stdin. Returns three arrays with identical length: file contents, table names, and context.
        The file names in the result may be different from the input file names, as STML files are replaced with their source.
        '''

        # we may have a single file coming from stdin, or multiple files from disk/sheet
        number_of_files = len(file_names) if sys.stdin.isatty() else 1

        # verify we have either as many tables and contexts, or none if we can derive from STML or file name
        assert table_names is None or len(table_names) == number_of_files, f"Provide exactly one table per file, or none. not {len(table_names)}"
        assert context is None or len(context) == number_of_files, f"Provide exactly one table per file, or none. not {len(context)}"

        # raise error if no contents are provided
        assert (file_names and len(file_names) > 0) or not sys.stdin.isatty(), 'No contents provided, either use --file or -f flag, or pipe data to stdin.'

        if not sys.stdin.isatty():
            assert file_names is None or len(file_names) == 0, 'Do not provide files using --file or -f when piping data to stdin.'
            # Input is being piped in, assume single file
            # read from stdin as binary
            file_contents = [sys.stdin.buffer.read()]
            file_names = ['stdin']
            context = context or ['stdin']

        else:
            # read all files
            file_contents = [open(file_name, 'rb').read() for file_name in file_names]
            # derive table names from file names if not provided
            table_names = table_names or [self._table_name_from_file_name(file_name) for file_name in file_names]
            # derive context from file names if not provided
            context = context or file_names

        # read as df, no header line, don't use nan
        file_contents_as_df = [pd.read_csv(io.BytesIO(file_content), header=None, keep_default_na=False) for file_content in file_contents]

        # instantiate STML evaluator with lambda to read file from disk
        stml_evaluator = StmlEvaluator(lambda original_file_name, source_file_name: self._read_file(original_file_name, source_file_name))

        # replace STML files with their source
        file_contents, table_names, context, substitutions = stml_evaluator.replace_stmls_with_sources(file_names, file_contents_as_df, table_names, context)

        return file_contents, table_names, context, substitutions

    def _table_name_from_file_name(self, file_name):
        # remove path and extension
        base_name = os.path.basename(file_name)
        # remove all characters after the first character that is not a letter or digit, or underscore
        return re.sub(r'[^a-zA-Z0-9_].*$', '', base_name)

    def _read_file(self, original_file_name, source_file_name):
        # assume the source file is in the same folder as the original file
        folder = os.path.dirname(original_file_name)
        # concat folder and source file name
        full_file_name = os.path.join(folder, source_file_name)
        # read file into dataframe
        return pd.read_csv(full_file_name)


class StmlEvaluator:

    def __init__(self, file_reader_lambda):
        # lambda to read source files, so we can use source files from file and from google sheets
        self._file_reader_lambda = file_reader_lambda

    def replace_stmls_with_sources(self, file_names, file_contents_as_df, table_names, context):
        # iterate over all files
        result = [self.replace_stml_with_source(n, f, t, c) for n, f, t, c in zip(file_names, file_contents_as_df, table_names, context)]
        # unzip result to file_contents, table_names, context, substitutions
        return zip(*result)

    def replace_stml_with_source(self, file_name, df, table_name, context):
        # check if cell A1 starts with '@'
        if not df.iloc[0, 0].startswith('@'):
            # not an STML file, return as binary CSV string
            csv = df.to_csv(index=False, header=False).encode('utf-8')
            return csv, table_name, context, None

        # get source file name, target table name and optional substitutions file name
        source_file_name = self._get_source_file_name(df, file_name)
        target_table_name = self._get_target_table_name(df, file_name)
        substitutions_file_name = self._get_substitutions_file_name(df, file_name)

        # read source and target columns from STML file
        source_and_target_column_list = self._get_source_and_target_columns(df, file_name)

        # use lambda to read source file into dataframe. Also pass original source file name to hint folder (if on disk)
        source_df = self._file_reader_lambda(file_name, source_file_name)

        # replace header line with target headers
        source_df = self._replace_header_line(source_and_target_column_list, source_df)

        # convert to csv
        source_csv = source_df.to_csv(index=False).encode('utf-8')

        # read substitutions file if the file name is provided
        if substitutions_file_name:
            substitutions_df = self._file_reader_lambda(file_name, substitutions_file_name)
            # convert to csv
            substitutions_csv = substitutions_df.to_csv(index=False).encode('utf-8')
        else:
            substitutions_csv = None

        # return source file as csv, target table name, the source file name and the substitions file as context
        return [source_csv, target_table_name, source_file_name, substitutions_csv]

    def _get_source_file_name(self, df, file_name):
        # assert it contains '@source'
        assert '@source' in df.iloc[0, 0], f"File {file_name} is an STML file, but does not contain '@source' in A1."
        # get source file name from B1
        source_file_name = df.iloc[0, 1]
        # assert it's not empty
        assert source_file_name, f"File {file_name} is an STML file, but B1 is empty."

        return source_file_name

    def _get_target_table_name(self, df, file_name):
        # assert A2 contains '@target'
        assert '@target' in df.iloc[1, 0], f"File {file_name} is an STML file, but does not contain '@target' in B2."
        # read target file name from B2
        target_table_name = df.iloc[1, 1]
        # assert it's not empty
        assert target_table_name, f"File {file_name} is an STML file, but B2 is empty."
        return target_table_name

    def _get_substitutions_file_name(self, df, file_name):
        # checkk if there's a substitutions header
        if not '@substitutions' in df.iloc[2, 0] or not df.iloc[2, 1]:
            # no substitutions file, return None
            return None

        # get substitutions file name from B3
        substitutions_file_name = df.iloc[2, 1]

        return substitutions_file_name

    def _get_source_and_target_columns(self, df, file_name):
        assert any(df[0] == 'source_column'), f"File {file_name} is an STML file, but does not contain 'source_column' in column A."
        # find first row number that contains 'source_column' in column A
        source_columns_row = df[df[0] == 'source_column'].index[0]
        # get modifier keys
        modifier_keys = self._get_modifier_column_names(df, source_columns_row)
        # create list of source and target columns, including modifiers
        column_rows = df.iloc[source_columns_row + 1:, :]
        # iterate over rows in dataframe using iterrows
        source_and_target_column_list = [self._get_source_and_target(row, modifier_keys) for _, row in column_rows.iterrows()]
        return source_and_target_column_list

    def _get_modifier_column_names(self, df, source_columns_row):
        # get list of additional non-empty headers to treat as modifiers from the source_columns_row
        modifiers = df.iloc[source_columns_row, 2:].tolist()
        # supported modifiers
        known_modifiers = ['unique', 'skip', 'default-value', 'exp', 'deduplicate', 'table', 'name', 'qualifier', 'key']
        # list unknown modifiers
        unknown_modifiers = [modifier for modifier in modifiers if modifier and modifier not in known_modifiers and modifier]
        # assert no unknown modifiers
        assert not unknown_modifiers, f'Unknown modifiers: {unknown_modifiers}'
        return modifiers

    def _get_source_and_target(self, row, modifiers_keys):
        source = row[0]
        target = row[1]
        # get columns 2 and up, with length equal to modifiers
        modifier_values = row[2:2 + len(modifiers_keys)]
        # create list of modifiers=value, where value is not empty
        modifier_list = list(map(lambda kv: kv[0] + '=' + kv[1], filter(lambda kv: kv[0] and kv[1], zip(modifiers_keys, modifier_values))))
        # create modifier string
        modifier_string = '[' + ': '.join(modifier_list) + ']' if modifier_list else ''
        # return source, target
        return source, target + modifier_string

    def _replace_header_line(self, source_and_target_column_list, source_df):
        # create mapping from source to target columns, excluding empty source columns
        source_to_target_mapping = {k: v for k, v in source_and_target_column_list if k}
        # create list of additional columns with empty source column to append to the end
        additional_columns = [v for k, v in source_and_target_column_list if not k]
        # replace headers with target headers
        source_df.columns = [source_to_target_mapping.get(c, '') for c in source_df.columns]
        # append additional columns to df
        source_df = pd.concat([source_df, pd.DataFrame(columns=additional_columns)], axis=1)
        return source_df
